# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import importlib.metadata
import os
import pathlib
from importlib.metadata import EntryPoint
from typing import Any, Callable, Union

import catalogue
import pytest
from composer.loggers import InMemoryLogger, LoggerDestination

from llmfoundry import registry
from llmfoundry.utils import registry_utils


def test_expected_registries_exist():
    existing_registries = {
        name for name in dir(registry)
        if isinstance(getattr(registry, name), registry_utils.TypedRegistry)
    }
    expected_registry_names = {
        'loggers',
        'optimizers',
        'schedulers',
        'tokenizers',
        'callbacks',
        'algorithms',
        'callbacks_with_config',
        'dataloaders',
        'dataset_replication_validators',
        'collators',
        'data_specs',
        'metrics',
        'models',
        'norms',
        'param_init_fns',
        'module_init_fns',
        'ffns',
        'ffns_with_norm',
        'ffns_with_megablocks',
        'attention_classes',
        'attention_implementations',
        'fcs',
        'icl_datasets',
        'config_transforms',
        'load_planners',
        'save_planners',
        'tp_strategies',
    }

    assert existing_registries == expected_registry_names


def test_registry_create(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(catalogue, 'Registry', {})

    new_registry = registry_utils.create_registry(
        'llmfoundry',
        'test_registry',
        generic_type=str,
        entry_points=False,
    )

    assert new_registry.namespace == ('llmfoundry', 'test_registry')
    assert isinstance(new_registry, registry_utils.TypedRegistry)


def test_registry_typing(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(catalogue, 'Registry', {})
    new_registry = registry_utils.create_registry(
        'llmfoundry',
        'test_registry',
        generic_type=str,
        entry_points=False,
    )
    new_registry.register('test_name', func='test')

    # This would fail type checking without the type ignore
    # It is here to show that the TypedRegistry is working (gives a type error without the ignore),
    # although this would not catch a regression in this regard
    new_registry.register('test_name', func=1)  # type: ignore


def test_registry_add(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(catalogue, 'Registry', {})
    new_registry = registry_utils.create_registry(
        'llmfoundry',
        'test_registry',
        generic_type=str,
        entry_points=False,
    )
    new_registry.register('test_name', func='test')

    assert new_registry.get('test_name') == 'test'


def test_registry_overwrite(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(catalogue, 'Registry', {})
    new_registry = registry_utils.create_registry(
        'llmfoundry',
        'test_registry',
        generic_type=str,
        entry_points=False,
    )
    new_registry.register('test_name', func='test')
    new_registry.register('test_name', func='test2')

    assert new_registry.get('test_name') == 'test2'


def test_registry_init_code(tmp_path: pathlib.Path):
    register_code = """
from llmfoundry.registry import loggers
from composer.loggers import InMemoryLogger

@loggers.register('test_logger')
class TestLogger(InMemoryLogger):
    pass
import os
os.environ['TEST_ENVIRON_REGISTRY_KEY'] = 'test'
"""

    with open(tmp_path / 'init_code.py', 'w') as _f:
        _f.write(register_code)

    registry_utils.import_file(tmp_path / 'init_code.py')

    assert issubclass(registry.loggers.get('test_logger'), InMemoryLogger)

    del catalogue.REGISTRY[('llmfoundry', 'loggers', 'test_logger')]

    assert 'test_logger' not in registry.loggers


def test_registry_entrypoint(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(catalogue, 'Registry', {})

    monkeypatch.setattr(
        importlib.metadata,
        'entry_points',
        lambda: {
            'llmfoundry_test_registry': [
                EntryPoint(
                    name='test_entry',
                    value='composer.loggers:InMemoryLogger',
                    group='llmfoundry_test_registry',
                ),
            ],
        },
    )

    monkeypatch.setattr(
        catalogue,
        'AVAILABLE_ENTRY_POINTS',
        importlib.metadata.entry_points(),
    )
    new_registry = registry_utils.create_registry(
        'llmfoundry',
        'test_registry',
        generic_type=str,
        entry_points=True,
    )
    assert new_registry.get('test_entry') == InMemoryLogger


def test_registry_builder(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setattr(catalogue, 'Registry', {})

    new_registry = registry_utils.create_registry(
        'llmfoundry',
        'test_registry',
        entry_points=False,
        generic_type=Union[type[LoggerDestination],
                           Callable[..., LoggerDestination]],
    )

    class TestLoggerDestination(LoggerDestination):
        pass

    new_registry.register('test_destination', func=TestLoggerDestination)

    # Valid, no validation
    valid_class = registry_utils.construct_from_registry(
        'test_destination',
        new_registry,
        pre_validation_function=TestLoggerDestination,
    )
    assert isinstance(valid_class, TestLoggerDestination)

    # Invalid, class validation
    with pytest.raises(
        ValueError,
        match='Expected test_destination to be of type',
    ):
        registry_utils.construct_from_registry(
            'test_destination',
            new_registry,
            pre_validation_function=InMemoryLogger,
        )

    # Invalid, function pre-validation
    with pytest.raises(ValueError, match='Invalid'):

        def pre_validation_function(x: Any):
            raise ValueError('Invalid')

        registry_utils.construct_from_registry(
            'test_destination',
            new_registry,
            pre_validation_function=pre_validation_function,
        )

    # Invalid, function post-validation
    with pytest.raises(ValueError, match='Invalid'):

        def post_validation_function(x: Any):
            raise ValueError('Invalid')

        registry_utils.construct_from_registry(
            'test_destination',
            new_registry,
            post_validation_function=post_validation_function,
        )

    # Invalid, not a class or function
    new_registry.register('non_callable', func=1)  # type: ignore
    with pytest.raises(
        ValueError,
        match='Expected non_callable to be a class or function',
    ):
        registry_utils.construct_from_registry('non_callable', new_registry)

    # Valid, partial function
    new_registry.register(
        'partial_func',
        func=lambda x,
        y: x * y,
    )  # type: ignore
    partial_func = registry_utils.construct_from_registry(
        'partial_func',
        new_registry,
        partial_function=True,
        kwargs={'x': 2},
    )
    assert partial_func(y=3) == 6

    # Valid, builder function
    new_registry.register('builder_func', func=lambda: TestLoggerDestination())
    valid_built_class = registry_utils.construct_from_registry(
        'builder_func',
        new_registry,
        partial_function=False,
    )
    assert isinstance(valid_built_class, TestLoggerDestination)
    assert os.environ['TEST_ENVIRON_REGISTRY_KEY'] == 'test'

    del os.environ['TEST_ENVIRON_REGISTRY_KEY']


def test_registry_init_code_fails(tmp_path: pathlib.Path):
    register_code = """
import os
os.environ['TEST_ENVIRON_REGISTRY_KEY'] = 'test'
asdf
"""

    with open(tmp_path / 'init_code.py', 'w') as _f:
        _f.write(register_code)

    with pytest.raises(RuntimeError, match='Error executing .*init_code.py'):
        registry_utils.import_file(tmp_path / 'init_code.py')


def test_registry_init_code_dne(tmp_path: pathlib.Path):
    with pytest.raises(FileNotFoundError, match='File .* does not exist'):
        registry_utils.import_file(tmp_path / 'init_code.py')



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Mapping

import torch
from composer.utils import dist
from transformers import AutoModel
from transformers.modeling_outputs import CausalLMOutputWithPast

from llmfoundry.models.llm_embed.modeling_llm_embed import ContrastiveModel


class FinetuneEmbeddingModel(ContrastiveModel):

    def construct_model(self) -> CausalLMOutputWithPast:
        # Define the model construction specific to FinetuneEmbeddingModel
        model = None

        def load_model():
            return AutoModel.from_pretrained(
                self.pretrained_model_name_or_path,
                trust_remote_code=self.trust_remote_code,
                use_auth_token=self.use_auth_token,
                **self.kwargs,
            )

        if dist.get_global_rank() == 0:
            model = load_model()
        dist.barrier()
        if model is None:
            model = load_model()

        assert model, 'Model is not loaded properly'
        return model

    def get_hidden_state(self, outputs: CausalLMOutputWithPast) -> torch.Tensor:
        """Override to return the last hidden state."""
        return outputs.last_hidden_state

    def handle_language_head(
        self,
        outputs: CausalLMOutputWithPast,
    ) -> torch.Tensor:
        """Override to skip language head handling."""
        return torch.tensor(
            0,
            dtype=torch.float32,
            device=outputs.last_hidden_state.device,
        )

    def flops_per_batch(self, batch: Mapping) -> int:
        # Get the batch size and maximum sequence length
        bs, msl = batch['input_ids'].shape[0:2]

        model_dimension = self._get_attribute(
            self.model.config,
            [
                'hidden_size',
                'd_model',
                'n_embd',
                'dim',
                'embed_dim',
                'embedding_size',
                'hidden_dim',
            ],
        )

        num_layers = self._get_attribute(
            self.model.config,
            [
                'num_hidden_layers',
                'n_layer',
                'num_layers',
                'encoder_layers',
                'decoder_layers',
                'n_layers',
                'num_blocks',
                'layer_count',
            ],
        )

        num_parameters = sum(p.numel() for p in self.model.parameters())

        # Estimate FLOPs
        params_flops = 2 * num_parameters
        seq_flops = params_flops * msl
        attn_flops = bs * num_layers * 2 * msl * model_dimension
        total_flops = seq_flops * bs + attn_flops

        return total_flops

    def _get_attribute(self, config: Any, possible_names: list):
        """Retrieve an attribute from config using a list of possible names."""
        for name in possible_names:
            value = getattr(config, name, None)
            if value is not None:
                return value
        return None



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from functools import lru_cache
from typing import Any, Optional

from transformers import PreTrainedTokenizer

__all__ = [
    'TiktokenTokenizerWrapper',
]

DEFAULT_SYSTEM_PROMPT = """You are a helpful, respectful and honest assistant. Always answer as helpfully as possible."""


# Taken from
# https://github.com/huggingface/transformers/blob/8aca43bdb3cb9a5020f6d57589d85679dc873b1c/src/transformers/models/gpt2/tokenization_gpt2.py#L62-L84
@lru_cache()
def bytes_to_unicode():
    """Returns list of utf-8 byte and a mapping to unicode strings.

    We specifically avoids mapping to whitespace/control characters the bpe code
    barfs on.

    The reversible bpe codes work on unicode strings. This means you need a
    large # of unicode characters in your vocab if you want to avoid UNKs. When
    you're at something like a 10B token dataset you end up needing around 5K
    for decent coverage. This is a significant percentage of your normal, say,
    32K bpe vocab. To avoid that, we want lookup tables between utf-8 bytes and
    unicode strings.
    """
    bs = (
        list(range(ord('!'),
                   ord('~') + 1)) + list(range(ord('¡'),
                                               ord('¬') + 1)) +
        list(range(ord('®'),
                   ord('ÿ') + 1))
    )
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))


class TiktokenTokenizerWrapper(PreTrainedTokenizer):
    """A thin wrapper around tiktoken to make it compatible with Hugging Face.

    tokenizers.

    See HuggingFace for further documentation on general tokenizer methods.
    """

    model_input_names = ['input_ids', 'attention_mask']

    def __init__(
        self,
        model_name: Optional[str] = None,
        encoding_name: Optional[str] = None,
        add_bos_token: bool = False,
        add_eos_token: bool = False,
        use_default_system_prompt: bool = False,
        unk_token: Optional[str] = '<|endoftext|>',
        eos_token: Optional[str] = '<|endoftext|>',
        bos_token: Optional[str] = '<|endoftext|>',
        pad_token: Optional[str] = None,
        errors: str = 'replace',
        chat_template: Optional[str] = None,
        **kwargs: Any,
    ):
        """Constructor creates a tiktoken tokenizer to use as the underlying.

        tokenizer.

        Args:
            model_name (Optional[str], optional): The name of the model to load from tiktoken. Defaults to None.
                Either model_name or encoding_name must be set, but not both.
            encoding_name (Optional[str], optional): The name of the encoding to load from tiktoken. Defaults to None.
                Either model_name or encoding_name must be set, but not both.
            add_bos_token (bool, optional): Whether to add bos tokens. Defaults to False.
            add_eos_token (bool, optional): Whether to add eos tokens. Defaults to False.
            use_default_system_prompt (bool, optional): Use the default system prompt or not. Defaults to False.
            unk_token (Optional[str], optional): The unk token. Defaults to '<|endoftext|>'.
            eos_token (Optional[str], optional): The eos token. Defaults to '<|endoftext|>'.
            bos_token (Optional[str], optional): The bos token. Defaults to '<|endoftext|>'.
            pad_token (Optional[str], optional): The pad token. Defaults to None.
            errors (str, optional): Paradigm to follow when decoding bytes to UTF-8. See
                [bytes.decode](https://docs.python.org/3/library/stdtypes.html#bytes.decode) for more information.
                Defaults to `"replace"`.
            chat_template (Optional[str], optional): The Hugging Face chat template. Default will use the ``default_chat_template``
                set on this class.
            kwargs (Any): Other relevant keyword arguments.
        """
        try:
            import tiktoken
        except:
            raise ImportError(
                'You need to install tiktoken to use TiktokenTokenizerWrapper.',
            )

        # Workaround to make tiktokenizer picklable.
        # https://github.com/huggingface/datasets/issues/5536#issuecomment-1682309347
        # There is an open PR from HF to add this to tiktoken: https://github.com/openai/tiktoken/pull/181
        import copyreg
        import functools

        from tiktoken import Encoding  # type: ignore (thirdParty)

        def pickle_Encoding(enc: Encoding):
            return (
                functools.partial(
                    Encoding,
                    enc.name,
                    pat_str=enc._pat_str,
                    mergeable_ranks=enc._mergeable_ranks,
                    special_tokens=enc._special_tokens,
                ),
                (),
            )

        copyreg.pickle(Encoding, pickle_Encoding)

        if model_name is not None and encoding_name is not None:
            raise ValueError(
                'You need to specify either model_name or encoding_name, not both.',
            )

        self.model_name = model_name
        self.encoding_name = encoding_name

        if self.model_name is not None:
            self.encoding = tiktoken.encoding_for_model(  # type: ignore (thirdParty)
                self.model_name)
        elif self.encoding_name is not None:
            self.encoding = tiktoken.get_encoding(  # type: ignore (thirdParty)
                self.encoding_name)
        else:
            raise ValueError(
                'You need to specify either model_name or encoding_name.',
            )

        self.add_bos_token = add_bos_token
        self.add_eos_token = add_eos_token
        self.use_default_system_prompt = use_default_system_prompt

        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        self.errors = errors

        self.decoder: dict[int, str] = {}
        for i in range(self.encoding.n_vocab):
            try:
                self.encoding.decode_single_token_bytes(i)
            except KeyError:
                continue
            # Taken from
            # https://gist.github.com/xenova/a452a6474428de0182b17605a98631ee
            decoding = ''.join([
                bytes_to_unicode()[ord(char)] for char in
                self.encoding.decode_single_token_bytes(i).decode('latin-1')
            ])
            self.decoder[i] = decoding

        self.encoder: dict[str, int] = {}
        for i in range(self.encoding.n_vocab):
            if i in self.decoder:
                self.encoder[self.decoder[i]] = i

        super().__init__(
            model_name=model_name,
            encoding_name=encoding_name,
            add_bos_token=add_bos_token,
            add_eos_token=add_eos_token,
            use_default_system_prompt=use_default_system_prompt,
            unk_token=unk_token,
            eos_token=eos_token,
            bos_token=bos_token,
            pad_token=pad_token,
            errors=errors,
            chat_template=chat_template or self.default_chat_template,
            **kwargs,
        )

    @property
    def vocab_size(self) -> int:
        """Returns vocab size."""
        return self.encoding.n_vocab

    @property
    def is_fast(self) -> bool:
        return False

    @property
    def default_chat_template(self):
        """Chat ML Template for User/Assistant.

        Pinning default Chat ML template in case defaults change.
        """
        template = (
            "{% if messages[0]['role'] == 'system' %}"
            '{% set loop_messages = messages[1:] %}'
            "{% set system_message = messages[0]['content'] %}"
            "{% elif USE_DEFAULT_PROMPT == true and not 'system' in messages[0]['role'] %}"
            '{% set loop_messages = messages %}'
            "{% set system_message = 'DEFAULT_SYSTEM_PROMPT' %}"
            '{% else %}'
            '{% set loop_messages = messages %}'
            '{% set system_message = false %}'
            '{% endif %}'
            '{% for message in loop_messages %}'
            '{% if loop.index0 == 0 %}'
            '{% if system_message != false %}'
            "{{ '<|im_start|>system\n' + system_message.strip() + '<|im_end|>\n'}}"
            '{% endif %}'
            "{{ '<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' }}"
            '{% else %}'
            "{{ '\n' + '<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' }}"
            '{% endif %}'
            '{% if (add_generation_prompt == true and loop.last) %}'
            "{{ '\n' + '<|im_start|>' + 'assistant' + '\n' }}"
            '{% endif %}'
            '{% endfor %}'
        )
        template = template.replace(
            'USE_DEFAULT_PROMPT',
            'true' if self.use_default_system_prompt else 'false',
        )
        template = template.replace(
            'DEFAULT_SYSTEM_PROMPT',
            DEFAULT_SYSTEM_PROMPT,
        )
        return template

    def get_vocab(self) -> dict[str, int]:
        """Returns vocab as a dict."""
        # As far as I can tell, we don't require get_vocab to completely work,
        # but when using additional_special_tokens, Hugging Face determines the next
        # token index to add with len(self.get_vocab()) so we need the _size_ of this dictionary to be correct.
        vocab_clone = self.encoder.copy()
        extra_id_index = 0
        candidate_extra_id = f'<extra_id_{extra_id_index}>'
        indices_to_fill_in = (
            set(range(self.vocab_size)) - set(vocab_clone.values())
        )

        # Add enough indices to make get_vocab() the right length
        for index_to_add in indices_to_fill_in:
            # Make sure we don't overwrite a token that already exists
            while candidate_extra_id in vocab_clone:
                extra_id_index += 1
                candidate_extra_id = f'<extra_id_{extra_id_index}>'

            # Get an index to add and add the item
            vocab_clone[candidate_extra_id] = index_to_add

        return dict(vocab_clone, **self.added_tokens_encoder)

    def _tokenize(self, text: str) -> list[str]:
        """Returns a tokenized string."""
        if not isinstance(text, str):
            raise ValueError(
                f'Expected a string input to _tokenize but got {type(text)}.',
            )

        tokens = [
            self.decoder[t]
            for t in self.encoding.encode(text, allowed_special='all')
        ]

        return tokens

    def _convert_token_to_id(self, token: str) -> Optional[int]:
        """Converts a token (str) in an id using the vocab."""
        return self.encoder.get(token, self.encoder.get(self.unk_token))

    def _convert_id_to_token(self, index: int) -> Optional[str]:
        """Converts an index (integer) in a token (str) using the vocab."""
        # For tokens in either the gap in ids in the tokenizer, or beyond the range of the tokenizer,
        # we return empty string. This matches the behavior of Hugging Face fast tokenizers,
        # but not slow tokenizers.
        return self.decoder.get(index, '')

    def convert_tokens_to_string(self, tokens: list[str]) -> str:
        """Converts a sequence of tokens (string) in a single string."""
        text = ''.join(tokens)
        text = bytearray([self.byte_decoder[c] for c in text
                         ],).decode('utf-8', errors=self.errors)
        return text

    def build_inputs_with_special_tokens(
        self,
        token_ids_0: list[int],
        token_ids_1: Optional[list[int]] = None,
    ) -> list[int]:
        bos_token_id = [self.bos_token_id] if self.add_bos_token else []
        eos_token_id = [self.eos_token_id] if self.add_eos_token else []

        output = bos_token_id + token_ids_0 + eos_token_id

        if token_ids_1 is not None:
            output = output + bos_token_id + token_ids_1 + eos_token_id

        return output

    def get_special_tokens_mask(
        self,
        token_ids_0: list[int],
        token_ids_1: Optional[list[int]] = None,
        already_has_special_tokens: bool = False,
    ) -> list[int]:
        """Retrieves sequence ids from a token list that has no special tokens.

        Function copied from
        https://github.com/huggingface/transformers/blob/e3a4bd2bee212a2d0fd9f03b27fe7bfc1debe42d/src/transformers/models/gpt2/tokenization_gpt2.py#L265-L295

        added. This method is called when adding special tokens using the
        tokenizer `prepare_for_model` or `encode_plus` methods.

        Args:
            token_ids_0 (`List[int]`):
                List of IDs.
            token_ids_1 (`List[int]`, *optional*):
                Optional second list of IDs for sequence pairs.
            already_has_special_tokens (`bool`, *optional*, defaults to `False`):
                Whether or not the token list is already formatted with special tokens for the model.

        Returns:
            `List[int]`: A list of integers in the range [0, 1]: 1 for a special token, 0 for a sequence token.
        """
        if already_has_special_tokens:
            return super().get_special_tokens_mask(
                token_ids_0=token_ids_0,
                token_ids_1=token_ids_1,
                already_has_special_tokens=True,
            )

        bos_token_id = [1] if self.add_bos_token else []
        eos_token_id = [1] if self.add_eos_token else []

        if token_ids_1 is None:
            return bos_token_id + ([0] * len(token_ids_0)) + eos_token_id
        return (
            bos_token_id + ([0] * len(token_ids_0)) + eos_token_id +
            bos_token_id + ([0] * len(token_ids_1)) + eos_token_id
        )

    def create_token_type_ids_from_sequences(
        self,
        token_ids_0: list[int],
        token_ids_1: Optional[list[int]] = None,
    ) -> list[int]:
        sep = [self.sep_token_id]

        if token_ids_1 is None:
            return len(token_ids_0 + sep) * [0]
        return len(token_ids_0 + sep) * [0] + len(token_ids_1 + sep) * [1]

    def save_vocabulary(
        self,
        save_directory: str,
        filename_prefix: Optional[str] = None,
    ) -> tuple[str]:

        # ignore the below type to keep the original signature
        # we are knowingly breaking the signature here, although not 100% certain
        # it doesn't have side effects
        # There is some code in huggingface that calls this function to get the vocab files,
        # but it doesn't seem to access them (or at least checks for their existence
        # before accessing them)
        return (None, None)  # type: ignore

    def sanitize_special_tokens(self) -> int:
        """Make sure that all the special tokens attributes of the tokenizer.

        (`tokenizer.mask_token`, `tokenizer.cls_token`, etc.) are in the
        vocabulary.

        Add the missing ones to the vocabulary if needed.

        Return:
            `int`: The number of tokens added in the vocabulary during the operation.
        """
        actual_new_tokens = []
        for token in self.all_special_tokens_extended:
            encoded = self.encoding.encode(token, allowed_special='all')
            if len(encoded) > 1:
                actual_new_tokens.append(token)

        return self.add_tokens(actual_new_tokens, special_tokens=True)


TiktokenTokenizerWrapper.register_for_auto_class()



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Dataset and dataloader for contrastive training.

Build a StreamingPairsDataset dataset and dataloader for contrastive training.
"""

import json
from itertools import islice
from typing import Any, Literal, Mapping, Optional, Union

import numpy as np
import torch
from composer.core import DataSpec
from composer.utils import retry
from streaming import Stream, StreamingDataset
from torch.utils.data import DataLoader
from transformers import PreTrainedTokenizerBase

from llmfoundry import registry
from llmfoundry.data.text_data import (
    ConcatenatedSequenceCollatorWrapper,
    StreamingTextDataset,
)
from llmfoundry.utils.registry_utils import construct_from_registry

ContrastiveSampleType = Literal['one_query_multiple_responses',
                                'one_query_one_response']


def _get_contrastive_sample_type(
    sample: Mapping[str, Any],
) -> ContrastiveSampleType:
    """Get the type of contrastive sample from the sample.

    Args:
        sample (Mapping): A sample from the dataset.

    Returns:
        ContrastiveSampleType: The type of contrastive sample.
    """
    sample_contains_text_a = any(
        key.startswith('text_a') for key in sample.keys()
    )
    sample_contains_text_b = any(
        key.startswith('text_b') for key in sample.keys()
    )

    if sample_contains_text_a and sample_contains_text_b:
        return 'one_query_one_response'
    elif 'query_text' in sample and 'positive_passage' in sample and 'negative_passages' in sample:
        return 'one_query_multiple_responses'
    else:
        raise ValueError(
            'Sample does not contain the required keys for contrastive training. \
                For datasets with one query and one response, the keys must contain \
                "text_a" and "text_b". For datasets with one query and multiple responses, \
                the keys must contain "query_text", "positive_passage", and "negative_passages".',
        )


class StreamingPairsDataset(StreamingTextDataset):
    """Contrastive pairs dataset using MosaicML's StreamingTextDataset.

    Args:
        max_hard_negatives (int, optional): The maximum number of hard negatives to include in the
            contrastive training samples. Defaults to ``None``. If ``None``, all hard negatives are
            included.
        prepend_query (str, optional): Text to prepend to the query text. Defaults to ``''``.
        prepend_passage (str, optional): Text to prepend to the passage text. Defaults to ``''``.
        append_eos_token (bool, optional): Whether to append the EOS token to the query and passage
            text. Defaults to ``False``. Mutually exclusive with ``append_token``.
        append_token (str, optional): Token to append to the query and passage text. Defaults to
            ``''``. Mutually exclusive with ``append_eos_token``.
        shuffle_hard_negatives (bool, optional): Whether to shuffle the hard negatives. Defaults to
            ``False``.
        **kwargs: Additional keyword arguments to pass to the superclass. See ``StreamingTextDataset``
            for more information.
    """

    def __init__(
        self,
        max_hard_negatives: Optional[int] = None,
        prepend_query: str = '',
        prepend_passage: str = '',
        append_eos_token: bool = False,
        append_token: str = '',
        shuffle_hard_negatives: bool = False,
        **kwargs: Any,
    ):

        super().__init__(**kwargs)

        self.max_hard_negatives = max_hard_negatives
        self.prepend_query = prepend_query
        self.prepend_passage = prepend_passage
        self.shuffle_hard_negatives = shuffle_hard_negatives
        self._generator = np.random.default_rng(seed=self.shuffle_seed)
        if append_eos_token:
            if append_token != '':
                raise ValueError(
                    'The arguments append_eos_token and append_token are mutually exclusive.',
                )
            self.append_token = self.tokenizer.eos_token
        else:
            self.append_token = append_token

    def _get_contrastive_samples(
        self,
        query_text: str,
        positive_response: str,
        negative_responses: list[str],
    ) -> dict[str, Union[str, list[str]]]:
        """Flatten contrastive samples into a list of strings.

        Args:
            query_text (str): The query text.
            positive_response (str): The positive response.
            negative_responses (List[str]): The negative responses.

        Returns:
            Dict[str, Union[str, List[str]]]: The contrastive samples, with keys 'query', 'positive', and 'negative'.
        """
        query_text = f'{self.prepend_query}{query_text}{self.append_token}'
        positive_response = f'{self.prepend_passage}{positive_response}{self.append_token}'
        if self.shuffle_hard_negatives:
            self._generator.shuffle(negative_responses)
        negative_responses = negative_responses[:self.max_hard_negatives]
        negative_responses = [
            f'{self.prepend_passage}{response}{self.append_token}'
            for response in negative_responses
        ]
        return {
            'query': query_text,
            'positive': positive_response,
            'negative': negative_responses,
        }

    @retry(BlockingIOError, num_attempts=5, initial_backoff=1.0, max_jitter=0.5)
    def __getitem__(self, idx: int) -> dict[str, list[int]]:
        sample = StreamingDataset.__getitem__(self, idx)
        text_samples = []

        sample_type = _get_contrastive_sample_type(sample)
        if sample_type == 'one_query_one_response':
            text_samples = self._get_contrastive_samples(
                sample['text_a'],
                sample['text_b'],
                [],
            )
        elif sample_type == 'one_query_multiple_responses':
            negative_passages_str = sample['negative_passages']
            text_samples = self._get_contrastive_samples(
                sample['query_text'],
                sample['positive_passage'],
                json.loads(negative_passages_str),
            )
        else:
            raise ValueError(f'Unknown sample type: {sample_type}')

        token_samples = self._tokenize(text_samples)
        return token_samples

    def _tokenize(
        self,
        text_samples: dict[str, Union[str, list[str]]],
    ) -> dict[str, list[int]]:
        if self.tokenizer.pad_token is None:
            raise RuntimeError(
                'If tokenizing on-the-fly, tokenizer must have a pad_token_id',
            )

        text_samples_list = [text_samples['query'], text_samples['positive']]
        text_samples_negatives = text_samples['negative']
        assert isinstance(text_samples_negatives, list)  # pyright type check
        text_samples_list.extend(text_samples_negatives)
        return self.tokenizer(
            text_samples_list,
            truncation=True,
            padding='max_length',
            max_length=self.max_seq_len,
        )


def build_pairs_dataloader(
    dataset: dict[str, Any],
    tokenizer: PreTrainedTokenizerBase,
    device_batch_size: int,
    drop_last: bool,
    num_workers: int,
    pin_memory: bool = True,
    prefetch_factor: int = 2,
    persistent_workers: bool = True,
    timeout: int = 0,
    max_hard_negatives: Optional[int] = None,
) -> DataSpec:
    dataset_cfg = dataset
    streams_dict = dataset.pop('streams', None)
    eos_token_id = dataset.pop('eos_token_id', None)
    bos_token_id = dataset.pop('bos_token_id', None)

    streams = None
    if streams_dict is not None:
        streams = []
        for stream in streams_dict.values():
            # stream is the streams kwargs
            # fwd all kwargs with **stream allows streaming to check args
            streams.append(Stream(**stream))

    pairs_dataset = StreamingPairsDataset(
        tokenizer=tokenizer,
        streams=streams,
        batch_size=device_batch_size,
        max_hard_negatives=max_hard_negatives,
        **dataset,
    )

    dataloader_cfg = {
        'name': 'contrastive_pairs',
        'dataset': dataset_cfg,
        'drop_last': drop_last,
        'num_workers': num_workers,
        'pin_memory': pin_memory,
        'prefetch_factor': prefetch_factor,
        'persistent_workers': persistent_workers,
        'timeout': timeout,
    }

    collate_fn, _ = construct_from_registry(
        name='text_collator',
        registry=registry.collators,
        partial_function=False,
        kwargs={
            'dataloader_cfg': dataloader_cfg,
            'tokenizer': tokenizer,
            'dataset_batch_size': device_batch_size,
        },
    )

    if (eos_token_id is not None) or (bos_token_id is not None):
        # Note: Will raise an error if both are non-None
        collate_fn = ConcatenatedSequenceCollatorWrapper(
            base_collator=collate_fn,
            eos_token_id=eos_token_id,
            bos_token_id=bos_token_id,
        )

    def collate_fn_without_labels(batch: list[Any]) -> dict[str, torch.Tensor]:
        # Contrastive learning does not require labels, with some embedding models even erroring out if they are present
        processed_batch: dict[str, torch.Tensor] = collate_fn(batch)
        if 'labels' in processed_batch:
            del processed_batch['labels']

        if 'total_tokens' in processed_batch:
            del processed_batch['total_tokens']
        if 'loss_generating_tokens' in processed_batch:
            del processed_batch['loss_generating_tokens']
        return processed_batch

    dl = DataLoader(
        pairs_dataset,
        collate_fn=collate_fn_without_labels,
        batch_size=device_batch_size,
        drop_last=drop_last,
        num_workers=num_workers,
        pin_memory=pin_memory,
        prefetch_factor=prefetch_factor,
        persistent_workers=persistent_workers,
        timeout=timeout,
    )

    return DataSpec(dataloader=dl)


# Helpful to test if your dataloader is working locally
# Run `python dataloader.py  --local_path [local] [--remote_path remote, optional]` and verify that batches are printed out
if __name__ == '__main__':
    import argparse

    from llmfoundry.utils.builders import build_tokenizer

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--tokenizer',
        type=str,
        default='EleutherAI/gpt-neox-20b',
        help='the name of the tokenizer to use',
    )
    parser.add_argument(
        '--local_path',
        type=str,
        required=True,
        help='the path to the local copy of the dataset',
    )
    parser.add_argument(
        '--remote_path',
        type=str,
        default=None,
        help='the path to the remote copy to stream from (optional)',
    )
    parser.add_argument(
        '--split',
        type=str,
        default='train',
        help='which split of the dataset to use',
    )
    parser.add_argument(
        '--max_seq_len',
        type=int,
        default=32,
        help='max sequence length to test',
    )

    args = parser.parse_args()

    if args.remote_path is not None:
        print(
            f'Reading {args.split} split from {args.local_path} <- streamed from <- {args.remote_path}',
        )
    else:
        print(f'Reading {args.split} split from {args.local_path}')

    cfg = {
        'name': 'contrastive_pairs',
        'dataset': {
            'local': args.local_path,
            'remote': args.remote_path,
            'split': args.split,
            'shuffle': False,
            'max_seq_len': args.max_seq_len,
            'keep_zip': True,  # in case we need compressed files after testing
        },
        'drop_last': False,
        'num_workers': 4,
    }
    device_batch_size = 2

    tokenizer_name = args.tokenizer
    tokenizer_kwargs = {'model_max_length': args.max_seq_len}
    tokenizer = build_tokenizer(tokenizer_name, tokenizer_kwargs)

    loader = build_pairs_dataloader(
        **cfg,
        tokenizer=tokenizer,
        device_batch_size=device_batch_size,
    ).dataloader
    assert isinstance(loader, DataLoader)
    assert isinstance(loader.dataset, StreamingPairsDataset)
    tokenizer = loader.dataset.tokenizer

    for batch_ix, batch in enumerate(islice(loader, 5)):
        print('\n')
        print('#' * 20, f'Batch {batch_ix}', '#' * 20)
        for k, v in batch.items():
            print(k, v.shape, v.dtype)
        for sample_ix, token_sample in enumerate(batch['input_ids']):
            print('-' * 20, f' Sample {sample_ix} ', '-' * 20)
            print(tokenizer.decode(token_sample))



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.data.contrastive_pairs.dataloader import (
    StreamingPairsDataset,
    build_pairs_dataloader,
)

__all__ = [
    'StreamingPairsDataset',
    'build_pairs_dataloader',
]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from functools import partial
from typing import Callable, Optional, Union

import torch
import torch.nn.functional as F

__all__ = [
    'dMoE',
    'LearnedRouter',
    'MLP',
    'GLU',
    'DroplessMLP',
]

DEFAULT_ACTIVATION_FN = partial(F.gelu, approximate='tanh')


# Add option to route tokens uniformly across experts. We use
# a custom autograd op router backwards is still run for benchmarking.
class _UniformExpertAssignment(torch.autograd.Function):

    @staticmethod
    def forward(
        ctx,  # pyright: ignore[reportMissingParameterType]
        x: torch.Tensor,
        num_experts: int,
    ):
        out = torch.arange(x.numel(), dtype=x.dtype, device=x.device)
        out = torch.remainder(out, num_experts)
        return out.view(x.shape)


class LearnedRouter(torch.nn.Module):

    def __init__(
        self,
        hidden_size: int,
        moe_num_experts: int,
        moe_top_k: int,
        moe_jitter_eps: Optional[float],
        moe_normalize_expert_weights: Optional[Union[int, float]],
        uniform_expert_assignment: bool,
        device: Optional[torch.device],
    ) -> None:
        super().__init__()
        self.hidden_size: int = hidden_size
        self.moe_num_experts: int = moe_num_experts
        self.moe_top_k: int = moe_top_k
        self.moe_jitter_eps: Optional[float] = moe_jitter_eps
        self.moe_normalize_expert_weights: Optional[Union[
            int, float]] = moe_normalize_expert_weights
        self.uniform_expert_assignment: bool = uniform_expert_assignment

        self.layer: torch.nn.Module = torch.nn.Linear(
            hidden_size,
            moe_num_experts,
            bias=False,
            device=device,
        )

    def jitter(self, x: torch.Tensor) -> torch.Tensor:
        assert self.moe_jitter_eps is not None
        low: float = 1.0 - self.moe_jitter_eps
        high: float = 1.0 + self.moe_jitter_eps
        noise: torch.Tensor = torch.rand(
            x.size(),
            dtype=x.dtype,
            device=x.device,
        )
        return low + noise * (high - low)

    def _top_k(self, scores: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        if self.moe_top_k == 1:
            values, indices = scores.max(dim=-1,)
            return values.unsqueeze(-1), indices.unsqueeze(-1)
        return torch.topk(
            scores,
            self.moe_top_k,
            dim=-1,
        )

    def forward(self, x: torch.Tensor):
        if self.training and self.moe_jitter_eps is not None:
            x = x * self.jitter(x)

        scores = self.layer(x.view(-1, x.shape[-1])).softmax(dim=-1)
        expert_weights, top_experts = self._top_k(scores)
        if self.moe_normalize_expert_weights:
            expert_weights = expert_weights / torch.norm(
                expert_weights,
                p=self.moe_normalize_expert_weights,
                dim=-1,
                keepdim=True,
            )

        top_experts = (
            _UniformExpertAssignment.apply(top_experts, self.moe_num_experts)
            if self.uniform_expert_assignment else top_experts
        )
        scores = scores.to(x.dtype)
        expert_weights = expert_weights.to(x.dtype)
        return scores, expert_weights, top_experts


class MLP(torch.nn.Module):

    def __init__(
        self,
        hidden_size: int,
        ffn_hidden_size: int,
        moe_num_experts: int,
        activation_fn: Callable,
        device: Optional[torch.device],
    ) -> None:
        super().__init__()

        self.moe_num_experts: int = moe_num_experts
        self.ffn_hidden_size: int = ffn_hidden_size
        self.hidden_size: int = hidden_size
        self.activation_fn: Callable = activation_fn

        self.w1 = torch.nn.Parameter(
            torch.rand(
                moe_num_experts * ffn_hidden_size,
                hidden_size,
                device=device,
            ),
        )
        self.w2 = torch.nn.Parameter(
            torch.rand(
                moe_num_experts * ffn_hidden_size,
                hidden_size,
                device=device,
            ),
        )
        self.activation_fn = activation_fn

    def forward(self, x: torch.Tensor, expert_idx: int) -> torch.Tensor:
        expert_w1 = self.w1.view(
            self.moe_num_experts,
            self.ffn_hidden_size,
            self.hidden_size,
        )[expert_idx]
        expert_w2 = self.w2.view(
            self.moe_num_experts,
            self.ffn_hidden_size,
            self.hidden_size,
        )[expert_idx]

        before_activation = x @ expert_w1.t()
        layer_1_output = self.activation_fn(before_activation)
        output = layer_1_output @ expert_w2
        return output


class GLU(torch.nn.Module):

    def __init__(
        self,
        hidden_size: int,
        ffn_hidden_size: int,
        moe_num_experts: int,
        activation_fn: Callable,
        device: Optional[torch.device],
    ):
        super().__init__()
        self.hidden_size = hidden_size
        self.ffn_hidden_size = ffn_hidden_size
        self.moe_num_experts = moe_num_experts

        self.w1 = torch.nn.Parameter(
            torch.rand(
                moe_num_experts * ffn_hidden_size,
                hidden_size,
                device=device,
            ),
        )
        self.v1 = torch.nn.Parameter(
            torch.rand(
                moe_num_experts * ffn_hidden_size,
                hidden_size,
                device=device,
            ),
        )
        self.w2 = torch.nn.Parameter(
            torch.rand(
                moe_num_experts * ffn_hidden_size,
                hidden_size,
                device=device,
            ),
        )
        self.activation_fn = activation_fn

    def forward(self, x: torch.Tensor, expert_idx: torch.Tensor):
        expert_w1 = self.w1.view(
            self.moe_num_experts,
            self.ffn_hidden_size,
            self.hidden_size,
        )[expert_idx]
        expert_v1 = self.v1.view(
            self.moe_num_experts,
            self.ffn_hidden_size,
            self.hidden_size,
        )[expert_idx]
        expert_w2 = self.w2.view(
            self.moe_num_experts,
            self.ffn_hidden_size,
            self.hidden_size,
        )[expert_idx]

        x1 = x.matmul(expert_w1.t())
        x2 = x.matmul(expert_v1.t())
        x1 = self.activation_fn(x1)
        x1 = x1 * x2
        x1 = x1.matmul(expert_w2)
        return x1


class DroplessMLP(torch.nn.Module):

    def __init__(
        self,
        hidden_size: int,
        ffn_hidden_size: int,
        mlp_type: str,
        moe_num_experts: int,
        activation_fn: Callable,
        bias: bool,
        device: Optional[torch.device],
    ):
        super().__init__()
        self.moe_num_experts = moe_num_experts

        if mlp_type == 'mlp':
            self.mlp = MLP(
                hidden_size=hidden_size,
                ffn_hidden_size=ffn_hidden_size,
                moe_num_experts=moe_num_experts,
                activation_fn=activation_fn,
                device=device,
            )
        elif mlp_type == 'glu':
            self.mlp = GLU(
                hidden_size=hidden_size,
                ffn_hidden_size=ffn_hidden_size,
                moe_num_experts=moe_num_experts,
                activation_fn=activation_fn,
                device=device,
            )
        else:
            raise ValueError(f'Received unknown {mlp_type=}')

    def forward(
        self,
        x: torch.Tensor,
        scores: torch.Tensor,
        expert_weights: torch.Tensor,
        top_experts: torch.Tensor,
    ):
        in_shape = x.shape
        hidden_size = in_shape[-1]

        x = x.view(-1, hidden_size)
        out = torch.zeros_like(x)

        expert_mask = torch.nn.functional.one_hot(
            top_experts,
            num_classes=self.moe_num_experts,
        ).permute(2, 1, 0)
        for expert_idx in range(0, self.moe_num_experts):
            topk_idx, token_idx = torch.where(expert_mask[expert_idx])
            if token_idx.shape[0] == 0:
                continue
            # In torch it is faster to index using lists than torch tensors
            token_list = token_idx.tolist()
            topk_list = topk_idx.tolist()

            expert_tokens = x[None, token_list].reshape(-1, hidden_size)
            mlp_output = self.mlp(expert_tokens, expert_idx)
            # In the following lines we move tensors to the same devices as the output of mlp. These operations should be no-ops during training.
            # This is done to fix pipeline parallel generation using hf.generate. Please see this comment for details: https://github.com/mosaicml/llm-foundry/pull/1332#issue-2386827204
            expert_weights = expert_weights.to(mlp_output.device)
            expert_out = mlp_output * expert_weights[token_list, topk_list,
                                                     None]
            out = out.to(mlp_output.device)
            token_idx = token_idx.to(mlp_output.device)
            out.index_add_(0, token_idx, expert_out)

        out = out.view(in_shape)
        return out


class dMoE(torch.nn.Module):

    def __init__(
        self,
        device: Optional[torch.device],
        hidden_size: int = 1024,
        ffn_hidden_size: int = 4096,
        moe_num_experts: int = 1,
        moe_top_k: int = 1,
        mlp_type: str = 'mlp',
        activation_fn: Callable = DEFAULT_ACTIVATION_FN,
        moe_jitter_eps: Optional[float] = None,
        moe_normalize_expert_weights: Optional[Union[int, float]] = None,
        uniform_expert_assignment: bool = False,
        bias: bool = True,
    ):
        super().__init__()

        # Token router.
        self.router = LearnedRouter(
            hidden_size,
            moe_num_experts=moe_num_experts,
            moe_top_k=moe_top_k,
            moe_jitter_eps=moe_jitter_eps,
            moe_normalize_expert_weights=moe_normalize_expert_weights,
            uniform_expert_assignment=uniform_expert_assignment,
            device=device,
        )

        # Expert computation helper.
        self.experts = DroplessMLP(
            hidden_size=hidden_size,
            ffn_hidden_size=ffn_hidden_size,
            mlp_type=mlp_type,
            moe_num_experts=moe_num_experts,
            activation_fn=activation_fn,
            bias=bias,
            device=device,
        )

    def forward(self, x: torch.Tensor):
        # Compute the expert scores and assignments.
        scores, expert_weights, top_experts = self.router(x)
        # Compute the experts.
        return self.experts(x, scores, expert_weights, top_experts)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Helper methods for the checkpoint conversion scripts.

The checkpoint conversion scripts are located in the
llmfoundry/scripts/inference/benchmarking/ folder. Users should run those
scripts directly to convert between checkpoints; this file contains only common
utility functions that are present in multiple scripts.
"""

import json
import logging
import os
import random
import string
from pathlib import Path
from typing import Any, Optional, Union

import numpy as np
from transformers import (
    AutoTokenizer,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)

log = logging.getLogger(__name__)

__all__ = [
    'get_hf_tokenizer_from_composer_state_dict',
    'load_tokenizer',
    'convert_and_save_ft_weights',
]


def _get_weight_data_type(data_type: str):
    if data_type == 'fp32':
        return np.float32
    elif data_type == 'fp16':
        return np.float16
    else:
        raise RuntimeError('Unsupported data type: {data_type} for conversion.')


# TODO: move this functionality to composer once the bug fixes are upstreamed
def get_hf_tokenizer_from_composer_state_dict(
    state_dict: dict[str, Any],
    trust_remote_code: bool,
    tokenizer_save_dir: Optional[str] = None,
) -> Optional[PreTrainedTokenizer]:
    if 'state' not in state_dict:
        raise RuntimeError(
            'Unexpected composer state dictionary. Did you pass in a full composer checkpoint?',
        )
    if 'integrations' not in state_dict[
        'state'] or 'huggingface' not in state_dict['state']['integrations']:
        raise RuntimeError(
            'Did not find HuggingFace related state (e.g., tokenizer) in the provided composer checkpoint!',
        )
    hf_tokenizer_state = state_dict['state']['integrations']['huggingface'][
        'tokenizer']
    hf_tokenizer = None
    if hf_tokenizer_state != {}:
        if tokenizer_save_dir is None:
            unique_suffix = ''.join(
                random.choices(string.ascii_letters + string.digits, k=6),
            )
            tokenizer_save_dir = os.path.join(
                os.getcwd(),
                f'tokenizer-save-dir-{unique_suffix}',
            )
        os.makedirs(tokenizer_save_dir, exist_ok=True)

        for filename, saved_content in hf_tokenizer_state.items():
            # For backwards compatibility, check if the filename already has the file extension
            if filename.endswith(saved_content['file_extension']):
                tokenizer_file_name = filename
            else:
                tokenizer_file_name = filename + saved_content['file_extension']

            # This cannot be a temporary directory because huggingface relies on the slow tokenizer file
            # being persistent on disk
            tokenizer_file_path = Path(tokenizer_save_dir) / tokenizer_file_name
            if saved_content['file_extension'] == '.json':
                with open(tokenizer_file_path, 'w') as _tmp_file:
                    json.dump(saved_content['content'], _tmp_file)
            elif saved_content['file_extension'] == '.txt':
                with open(tokenizer_file_path, 'w') as _tmp_file:
                    for line in saved_content['content']:
                        _tmp_file.write(line)
                        _tmp_file.write('\n')
            elif saved_content['file_extension'] == '.py':
                with open(tokenizer_file_path, 'w') as _tmp_file:
                    _tmp_file.write(saved_content['content'])
            elif saved_content['file_extension'] == '.model':
                try:
                    import sentencepiece as spm
                except ImportError as e:
                    raise ImportError(
                        'Your tokenizer uses `sentencepiece`. Please install `sentencepiece` to load it.',
                    ) from e

                s = spm.SentencePieceProcessor()
                s.load_from_serialized_proto(saved_content['content'])
                with open(tokenizer_file_path, 'wb') as _tmp_file:
                    _tmp_file.write(s.serialized_model_proto())

        hf_tokenizer = load_tokenizer(
            tokenizer_save_dir,
            trust_remote_code=trust_remote_code,
        )

        # remove 'name_or_path'
        hf_tokenizer.name_or_path = ''
        hf_tokenizer.init_kwargs['name_or_path'] = ''

    return hf_tokenizer


def load_tokenizer(
    tokenizer_save_dir: str,
    trust_remote_code: bool,
) -> Union[PreTrainedTokenizer, PreTrainedTokenizerFast]:
    try:
        return AutoTokenizer.from_pretrained(
            tokenizer_save_dir,
            trust_remote_code=trust_remote_code,
        )
    except ValueError as e:
        raise ValueError(
            f'Got error while loading tokenizer with trust_remote_code={trust_remote_code}: {e}. '
            +
            'If accessing a tokenizer defined outside of the transformers module,'
            + ' please use --trust_remote_code.',
        )


def _write_zero_bias(
    weight_name: str,
    weight_file_path: str,
    bias_shape: Union[tuple[int, ...], int],
    np_data_type: np.dtype,
) -> None:
    """Write zeros for bias when converting MPT to FasterTransformer weights.

    MPT model might not have bias while FT expects bias.

    Args:
        weight_name (str): Name of the weight tensor.
        weight_file_path (str): Output path for storing the weight (NOT zero bias).
        bias_shape (Union[Tuple[int, ...], int]): Shape of the bias array.
        np_data_type (np.dtype): The data type for bias.
    """
    if 'weight' not in weight_file_path:
        raise RuntimeError(
            f'Cannot write zero bias for {weight_name}. Input is not a weight tensor',
        )
    log.debug(f'zero bias for weight: {weight_name}')
    bias_file_path = weight_file_path.replace('.weight', '.bias')
    bias = np.zeros(bias_shape, dtype=np_data_type)
    bias.tofile(bias_file_path)


def _convert_weight_to_ft_each(
    save_dir: str,
    infer_gpu_num: int,
    tensor_name: str,
    config: dict[str, Any],
    data: np.ndarray,
    np_weight_data_type: np.dtype,
) -> None:
    """Convert each MPT weight to a FasterTransformer compatible format.

    Args:
        save_dir (str): Path of the directory to save the weight in FT format. The directory must already exist.
        infer_gpu_num (int): The number of gpus you are planning to use for inference.
        tensor_name (str): Name of the weight tensor. Used in naming the output file.
        config (Dict[str, Any]): Configuration for the model. This is used in getting model specific parameters.
        data (np.ndarray): Tensor data in np.ndarray format.
        np_weight_data_type (np.dtype): Data type of the numpy array `data`.

    Returns:
        None: Writes to a file in `save_dir`. File name is based on the `tensor_name`
    """
    if tensor_name.find('input_layernorm.weight') != -1 or tensor_name.find('input_layernorm.bias') != -1 or \
        tensor_name.find('attention.dense.bias') != -1 or tensor_name.find('post_attention_layernorm.weight') != -1 or \
        tensor_name.find('post_attention_layernorm.bias') != -1 or tensor_name.find('mlp.dense_4h_to_h.bias') != -1 or \
        tensor_name.find('final_layernorm.weight') != -1 or tensor_name.find('final_layernorm.bias') != -1:

        save_path = os.path.join(save_dir, f'model.{tensor_name}.bin')
        data.tofile(save_path)
        if 'weight' in tensor_name and config['no_bias']:
            _write_zero_bias(
                tensor_name,
                save_path,
                data.shape[-1],
                np_weight_data_type,
            )  # pyright: ignore [reportGeneralTypeIssues]

    elif tensor_name.find('attention.dense.weight') != -1:
        assert data.shape == (
            config['d_model'],
            config['d_model'],
        ), f'unexpected dim for {tensor_name}'
        # nn.Linear weights are transposed
        data = data.T
        split_vals = np.split(data, infer_gpu_num, axis=0)
        for j in range(infer_gpu_num):
            save_path = os.path.join(save_dir, f'model.{tensor_name}.{j}.bin')
            split_vals[j].tofile(save_path)
        if config['no_bias']:
            fake_weight_path = os.path.join(
                save_dir,
                f'model.{tensor_name}.bin',
            )
            _write_zero_bias(
                tensor_name,
                fake_weight_path,
                data.shape[-1],
                np_weight_data_type,
            )  # pyright: ignore [reportGeneralTypeIssues]

    elif tensor_name.find('mlp.dense_4h_to_h.weight') != -1:
        assert data.shape == (
            config['d_model'],
            config['expansion_ratio'] * config['d_model'],
        ), f'unexpected dim for {tensor_name}'
        # nn.Linear weights are transposed
        data = data.T
        split_vals = np.split(data, infer_gpu_num, axis=0)
        for j in range(infer_gpu_num):
            save_path = os.path.join(save_dir, f'model.{tensor_name}.{j}.bin')
            split_vals[j].tofile(save_path)
        if config['no_bias']:
            fake_weight_path = os.path.join(
                save_dir,
                f'model.{tensor_name}.bin',
            )
            _write_zero_bias(
                tensor_name,
                fake_weight_path,
                data.shape[-1],
                np_weight_data_type,
            )  # pyright: ignore [reportGeneralTypeIssues]

    elif tensor_name.find('mlp.dense_h_to_4h.weight') != -1:
        assert data.shape == (
            config['expansion_ratio'] * config['d_model'],
            config['d_model'],
        ), f'unexpected dim for {tensor_name}'
        # nn.Linear weights are transposed
        data = data.T

        split_vals = np.split(data, infer_gpu_num, axis=-1)
        for j in range(infer_gpu_num):
            save_path = os.path.join(save_dir, f'model.{tensor_name}.{j}.bin')
            split_vals[j].tofile(save_path)
            if config['no_bias']:
                _write_zero_bias(
                    tensor_name,
                    save_path,
                    split_vals[j].shape[-1],
                    np_weight_data_type,
                )  # pyright: ignore [reportGeneralTypeIssues]

    elif tensor_name.find('mlp.dense_h_to_4h.bias') != -1:
        assert data.shape == (
            config['expansion_ratio'] * config['d_model'],
        ), f'unexpected dim for {tensor_name}'
        split_vals = np.split(data, infer_gpu_num, axis=-1)
        for j in range(infer_gpu_num):
            save_path = os.path.join(save_dir + f'model.{tensor_name}.{j}.bin')
            split_vals[j].tofile(save_path)

    elif tensor_name.find('attention.query_key_value.bias') != -1:
        assert data.shape == (
            3 * config['d_model'],
        ), f'unexpected dim for {tensor_name}'

        data = data.reshape(3, config['d_model'])

        split_vals = np.split(data, infer_gpu_num, axis=-1)

        for j in range(infer_gpu_num):
            save_path = os.path.join(save_dir, f'model.{tensor_name}.{j}.bin')
            split_vals[j].tofile(save_path)

    elif tensor_name.find('attention.query_key_value.weight') != -1:
        assert data.shape == (
            3 * config['d_model'],
            config['d_model'],
        ), f'unexpected dim for {tensor_name}'
        # nn.Linear weights are transposed
        data = data.T

        data = data.reshape(config['d_model'], 3, config['d_model'])
        split_vals = np.split(data, infer_gpu_num, axis=-1)

        for j in range(infer_gpu_num):
            save_path = os.path.join(save_dir, f'model.{tensor_name}.{j}.bin')
            split_vals[j].tofile(save_path)
            if config['no_bias']:
                _write_zero_bias(
                    tensor_name,
                    save_path,
                    (3, split_vals[j].shape[-1]),
                    np_weight_data_type,
                )  # pyright: ignore [reportGeneralTypeIssues]

    else:
        raise RuntimeError(f'Tensor with name {tensor_name} is not handled')


def convert_and_save_ft_weights(
    named_params: dict,
    config: dict,
    infer_gpu_num: int = 1,
    weight_data_type: str = 'fp32',
    save_dir: str = '',
) -> None:
    """Convert a Composer MPT checkpoint to a FasterTransformer format.

    Args:
        named_params (Dict[str, Parameter]): A dictionary containing the Composer MPT model's parameter names and data.
        config (Dict[str, Any]): Configuration for the model. This is used in getting model specific parameters.
        infer_gpu_num (int): The number of gpus you are planning to use for inference.
        weight_data_type (str): The dtype of the converted FasterTransformer model.
        save_dir (str): Path of the directory to save the weight in FT format. The directory must already exist.

    Returns:
        None: Writes to the `save_dir` folder. File names within this folder are based on the model parameter names.
    """
    np_weight_data_type = _get_weight_data_type(weight_data_type)

    param_remapping = {
        'norm_1.bias': 'input_layernorm.bias',
        'norm_1.weight': 'input_layernorm.weight',
        'attn.Wqkv.bias': 'attention.query_key_value.bias',
        'attn.Wqkv.weight': 'attention.query_key_value.weight',
        'attn.out_proj.bias': 'attention.dense.bias',
        'attn.out_proj.weight': 'attention.dense.weight',
        'norm_2.bias': 'post_attention_layernorm.bias',
        'norm_2.weight': 'post_attention_layernorm.weight',
        'ffn.up_proj.bias': 'mlp.dense_h_to_4h.bias',
        'ffn.up_proj.weight': 'mlp.dense_h_to_4h.weight',
        'ffn.down_proj.bias': 'mlp.dense_4h_to_h.bias',
        'ffn.down_proj.weight': 'mlp.dense_4h_to_h.weight',
    }

    for name, param in named_params.items():
        log.debug(f'Working on parameter {name} ...')
        data = param.detach().cpu().numpy().astype(np_weight_data_type)
        if name.find('weight') == -1 and name.find('bias') == -1:
            log.debug(f'found a parameter name that is not handled: {name}')
            continue
        if name == 'transformer.wpe.weight':
            assert data.shape == (
                config['max_seq_len'],
                config['d_model'],
            ), f'unexpected dim for {name}'
            data.tofile(os.path.join(save_dir, 'model.wpe.bin'))
        elif name == 'transformer.wte.weight':
            assert data.shape == (
                config['vocab_size'],
                config['d_model'],
            ), f'unexpected dim for {name}'
            data.tofile(os.path.join(save_dir, 'model.wte.bin'))
        elif name == 'transformer.norm_f.bias':
            assert data.shape == (
                config['d_model'],
            ), f'unexpected dim for {name}'
            data.tofile(
                os.path.join(save_dir, 'model.final_layernorm.bias.bin'),
            )
        elif name == 'transformer.norm_f.weight':
            assert data.shape == (
                config['d_model'],
            ), f'unexpected dim for {name}'
            save_path = os.path.join(
                save_dir,
                'model.final_layernorm.weight.bin',
            )
            data.tofile(save_path)
            if config['no_bias']:
                _write_zero_bias(
                    name,
                    save_path,
                    data.shape[-1],
                    np_weight_data_type,  # pyright: ignore [reportGeneralTypeIssues]
                )
        elif name == 'transformer.lm_head.weight':
            data.tofile(os.path.join(save_dir, 'model.lm_head.weight.bin'))
        else:
            for mpt_pattern, ft_pattern in param_remapping.items():
                if name.find(mpt_pattern) != -1:
                    new_name = name.replace(
                        'transformer.blocks.',
                        'layers.',
                    ).replace(mpt_pattern, ft_pattern)
                    _convert_weight_to_ft_each(
                        save_dir,
                        infer_gpu_num,
                        new_name,
                        config,
                        data,
                        np_weight_data_type,  # pyright: ignore [reportGeneralTypeIssues]
                    )



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Helper function to configure MPT with MoEs."""

import inspect
from typing import Callable, Optional, Union

import torch
from packaging import version
from torch import distributed
from torch.distributed._tensor import DeviceMesh

from llmfoundry.layers_registry import ffns_with_megablocks
from llmfoundry.models.layers.ffn import resolve_ffn_hidden_size

__all__ = [
    'config_moe_args',
]


def create_process_group_ranks(ranks: tuple[int, ...]):
    """Creates a new distributed group.

    Used in create_set_process_group and create_mod_process_group methods below.

    This function is an alternative to `distributed.new_group(ranks)`.

    Args:
        ranks (tuple[int, ...]): Tuple of ranks of group members.

    Returns:
        A handle of distributed group that can be given to collective calls.
    """
    ranks_gather_list = [None for _ in range(distributed.get_world_size())]
    distributed.all_gather_object(ranks_gather_list, ranks)
    ranks_per_subgroup = list(set(ranks_gather_list))
    group, _ = distributed.distributed_c10d.new_subgroups_by_enumeration(
        ranks_per_subgroup,
    )
    return group


def create_set_process_group(k: int):
    """Creates a new distributed group using sets of k GPUs.

    For example, if you have 16 GPUs and input k=4, the resulting process groups
    will have ranks:
        process group 0 ranks: [ 0,  1,  2,  3]
        process group 1 ranks: [ 4,  5,  6,  7]
        process group 2 ranks: [ 8,  9, 10, 11]
        process group 3 ranks: [12, 13, 14, 15]

    Args:
        k (int): Number of GPUs to use in set size.

    Returns:
        A handle of distributed group that can be given to collective calls.
    """
    world_size = distributed.get_world_size()
    if world_size % k != 0:
        raise RuntimeError(f'{world_size=} must be divisible by {k=}.')
    start = distributed.get_rank() // k * k
    ranks = tuple(range(start, start + k))
    return create_process_group_ranks(ranks)


def get_megablocks_device_mesh(
    device_mesh_cfg: Optional[tuple[int, ...]],
    moe_world_size: int,
    world_size: int,
) -> DeviceMesh:
    """Helper function to get the device mesh for MegaBlocks MoE.

    Args:
        device_mesh_cfg (Optional[tuple[int, ...]]): The device mesh configuration specification.
        moe_world_size (int): The MoE world size.
        world_size (int): The world size.

    Raises:
        ValueError: If the device mesh configuration is not valid.

    Returns:
        The device mesh for MegaBlocks MoE.
    """
    from torch.distributed._tensor.device_mesh import init_device_mesh

    if device_mesh_cfg is None or len(device_mesh_cfg) == 1:
        if device_mesh_cfg is not None:
            world_size = device_mesh_cfg[0]
        sharding_group_dim = world_size // moe_world_size
        device_mesh = init_device_mesh(
            'cuda',
            (sharding_group_dim, moe_world_size),
            mesh_dim_names=('weight_parallel', 'expert_parallel'),
        )
    else:
        raise ValueError(f'{device_mesh_cfg=} must be length 1')

    return device_mesh


def config_megablocks_moe_args(
    ffn_config: dict,
    d_model: int,
    expansion_ratio: Union[int, float],
    n_layers: int,
    get_device_mesh: Callable,
) -> dict:
    """Configures `ffn_config` for MegaBlocks MoE.

    We prepare all necessary arguments for `megablocks.layers.arguments.Arguments` so that process
    groups can be initialized and shared across all blocks in the network.

    Args:
        ffn_config (dict): FFN configuration before the MegaBlocks MoE is configured.
        d_model (int): Hidden size of the network.
        expansion_ratio (Union[int, float]): Expansion ratio in FFN.
        n_layers (int): Number of blocks used in the network.
        get_device_mesh (Callable): Function to get the device mesh. Takes in the device mesh config and the MoE world size.

    Returns:
        ffn_config (dict): FFN configuration with MegaBlocks MoE configured.
    """
    try:
        import megablocks
    except:
        raise RuntimeError(
            'Requirements for MegaBlocks not installed; see install instructions in `README.md`.',
        )

    ffn_config.setdefault('fp16', False)
    ffn_config.setdefault('bf16', False)
    ffn_config['num_layers'] = n_layers

    ffn_type = ffn_config.pop('ffn_type')
    fc_type = ffn_config.pop('fc_type')
    ffn_act_fn = ffn_config.pop('ffn_act_fn', None)

    # Config for MegaBlocks MoE world size and device mesh
    world_size = 1  # default
    moe_world_size = ffn_config.pop('moe_world_size')
    device_mesh = None
    device_mesh_cfg = ffn_config.pop('device_mesh', None)
    if moe_world_size > 1:
        if version.parse(
            torch.__version__.split('.dev')[0],
        ) < version.parse('2.2.0'):  # type: ignore
            raise RuntimeError(
                'MoE world size > 1 is not supported in torch version {torch.__version__}<2.2.',
            )

        world_size = distributed.get_world_size()
        if world_size < moe_world_size or world_size % moe_world_size:
            raise ValueError(
                f'Invalid world size configuration: {world_size=} and {moe_world_size=}',
            )

        device_mesh = get_device_mesh(
            device_mesh_cfg=device_mesh_cfg,
            moe_world_size=moe_world_size,
            world_size=world_size,
        )

        ffn_config['moe_expert_model_parallelism'] = True
        ffn_config['expert_parallel_group'] = device_mesh[
            'expert_parallel'].get_group(0)  # type: ignore

    lbl_process_group = ffn_config.get('lbl_process_group', None)
    if lbl_process_group is not None:
        if lbl_process_group == 'expert_group':
            lbl_process_group = ffn_config['expert_parallel_group']
        elif lbl_process_group == 'global_group':
            lbl_process_group = distributed.group.WORLD
        elif isinstance(lbl_process_group, int):
            if lbl_process_group > 1:
                lbl_process_group = create_set_process_group(lbl_process_group)
            else:
                lbl_process_group = None
        elif not isinstance(lbl_process_group, distributed.ProcessGroup):
            raise ValueError(
                f'Unknown {lbl_process_group=}. Options are: none | a process group | ``expert_group`` | ``global_group`` | <GROUP_SIZE>.',
            )
        ffn_config['lbl_process_group'] = lbl_process_group

    ffn_hidden_size = resolve_ffn_hidden_size(d_model, expansion_ratio)
    ffn_config.setdefault('ffn_hidden_size', ffn_hidden_size)

    args_to_keep_in_ffn_config = inspect.signature(
        megablocks.layers.arguments.Arguments,
    ).parameters

    ffn_config = {
        k: v for k, v in ffn_config.items() if k in args_to_keep_in_ffn_config
    }

    args = megablocks.layers.arguments.Arguments(
        hidden_size=d_model,
        **ffn_config,
    )
    ffn_config['args'] = args
    ffn_config['device_mesh'] = device_mesh
    ffn_config['moe_world_size'] = moe_world_size
    ffn_config['ffn_type'] = ffn_type
    ffn_config['fc_type'] = fc_type
    ffn_config['ffn_act_fn'] = ffn_act_fn

    return ffn_config


def config_moe_args(
    ffn_config: dict,
    d_model: int,
    expansion_ratio: Union[int, float],
    n_layers: int,
) -> dict:
    """Configures `ffn_config` for MoE.

    Args:
        ffn_config (dict): FFN configuration before the MoE is configured.
        d_model (int): Hidden size of the network.
        expansion_ratio (int, float): Expansion ratio in FFN.
        n_layers (int): Number of blocks used in the network.

    Returns:
        ffn_config (dict): FFN configuration with MoE configured.
    """
    if ffn_config['ffn_type'] in ffns_with_megablocks:
        return config_megablocks_moe_args(
            ffn_config=ffn_config,
            d_model=d_model,
            expansion_ratio=expansion_ratio,
            n_layers=n_layers,
            get_device_mesh=get_megablocks_device_mesh,
        )
    else:
        raise ValueError(f'Invalid ffn_type ({ffn_config["ffn_type"]}).')



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import contextlib
import os
import tempfile
from pathlib import Path

import torch
from composer.core import Callback, State
from composer.core.state import (
    fsdp_get_optim_state_dict,
    fsdp_state_dict_type_context,
)
from composer.loggers import Logger
from composer.loggers.remote_uploader_downloader import RemoteUploaderDownloader
from composer.utils import (
    dist,
    format_name_with_dist_and_time,
    parse_uri,
    reproducibility,
)

__all__ = ['MonolithicCheckpointSaver']


class MonolithicCheckpointSaver(Callback):
    """Save a monolithic checkpoint every N batches.

    Args:
        save_folder (str): Folder to save checkpoints to (can be a URI)
        batch_interval (int): Number of batches between checkpoints.
        filename (str): Filename to save checkpoints to.
        overwrite (bool): Whether to overwrite previous checkpoints.
        keep_optimizers (bool): Whether to save the optimizer state in the monolithic checkpoint.
    """

    def __init__(
        self,
        save_folder: str,
        batch_interval: int,
        filename: str = 'ep{epoch}-ba{batch}.pt',
        overwrite: bool = False,
        keep_optimizers: bool = False,
    ):
        self.backend, self.bucket_name, self.save_dir_format_str = parse_uri(
            save_folder,
        )
        self.filename_format_str = filename
        self.batch_interval = batch_interval
        self.upload_to_object_store = (self.backend != '')
        self.overwrite = overwrite
        self.keep_optimizers = keep_optimizers
        if self.upload_to_object_store:
            self.remote_ud = RemoteUploaderDownloader(
                bucket_uri=f'{self.backend}://{self.bucket_name}',
            )
        else:
            self.remote_ud = None

    def init(self, state: State, logger: Logger) -> None:
        if self.upload_to_object_store and self.remote_ud is not None:
            self.remote_ud.init(state, logger)
            # updated_logger_destinations = [*logger.destinations, new_remote_ud]
            # logger.destinations = tuple(updated_logger_destinations)
            state.callbacks.append(self.remote_ud)

    def batch_checkpoint(self, state: State, logger: Logger) -> None:
        if state.timestamp.batch.value % self.batch_interval == 0:
            self._save_checkpoint(state, logger)

    def fit_end(self, state: State, logger: Logger) -> None:
        if state.timestamp.batch.value % self.batch_interval != 0:
            self._save_checkpoint(state, logger)

    def _save_checkpoint(self, state: State, logger: Logger) -> None:
        del logger  # unused

        filename = format_name_with_dist_and_time(
            self.filename_format_str,
            state.run_name,
            state.timestamp,
        )
        save_dir = format_name_with_dist_and_time(
            self.save_dir_format_str,
            state.run_name,
            state.timestamp,
        )
        dir_context_mgr = tempfile.TemporaryDirectory(
        ) if self.upload_to_object_store else contextlib.nullcontext(
            enter_result=save_dir,
        )
        with dir_context_mgr as temp_save_dir:
            # pyright doesn't know about enter_result
            assert isinstance(temp_save_dir, str)

            save_path = str(Path(temp_save_dir) / Path(filename))
            dirname = os.path.dirname(save_path)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            state_dict = {
                'state': state.state_dict(),
                'rng': reproducibility.get_rng_state(),
            }
            # Remove sharded model and optimizer state dicts
            state_dict['state'].pop('optimizers')
            state_dict['state'].pop('model')

            # Add in unsharded model params.
            with fsdp_state_dict_type_context(
                state.model,
                state_dict_type='full',
            ):
                state_dict['state']['model'] = state.model.state_dict()

            # Add in unsharded optimizer state dict.
            if self.keep_optimizers:
                optimizer = state.optimizers[0]
                state_dict['state']['optimizers'] = {
                    type(optimizer).__qualname__:
                        fsdp_get_optim_state_dict(
                            state.model,
                            optimizer,
                            state_dict_type='full',
                        ),
                }
            if dist.get_global_rank() == 0:
                torch.save(state_dict, save_path)

            if self.upload_to_object_store and self.remote_ud is not None and dist.get_global_rank(
            ) == 0:
                remote_file_name = str(Path(save_dir) / Path(filename))
                self.remote_ud.upload_file(
                    state=state,
                    remote_file_name=remote_file_name,
                    file_path=Path(save_path),
                    overwrite=self.overwrite,
                )



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Enable curriculum learning by specifying a schedule of datasets to train on.

This module provides a CurriculumLearning callback that allows for dynamic
dataset switching during training based on a predefined schedule.
"""

import copy
import logging
from dataclasses import dataclass
from typing import Any, Union

from composer import DataSpec
from composer.core import State, Time, TimeUnit, ensure_time
from composer.loggers import Logger
from streaming import StreamingDataset
from streaming.base.util import clean_stale_shared_memory
from torch.utils.data import DataLoader

from llmfoundry.interfaces import CallbackWithConfig
from llmfoundry.utils.exceptions import (
    BaseContextualError,
    TrainDataLoaderLocation,
)

log = logging.getLogger(__name__)

__all__ = ['CurriculumLearning']


@dataclass
class CurriculumLearningState:
    schedule: list[dict[str, Any]]
    schedule_index: int


class CurriculumLearning(CallbackWithConfig):
    """Starts an epoch with a different dataset when resuming from a checkpoint.

    Example duration:
    <number>tok
    Example schedule:
    [
        {
            'duration': <number>tok,
            'dataset': <dataset parameters>,
        },
        {
            'duration': <number>tok,
            'dataset': <dataset parameters>,
        },
        {
            'duration': <number>tok,
            'dataset': <dataset parameters>,
        ],
    ]

    Args:
        train_config (dict): The configuration of the dataset currently
            being used. Note that this is the full train config and must
            contain the 'train_loader', 'device_train_batch_size', and
            'tokenizer' keys.
        duration (Union[Time, str, int]): The duration of the first datamix
            (which corresponds to the train_loader).
        schedule (list[dict[str, Any]]): The list of datamixes to use and their
            durations. Duration units must match max_duration and be in terms of
            a TimeUnit that is supported by Iteration. The duration values must
            be positive. There must be at least one datamix in the schedule. The
            first datamix during training is not included in the schedule. On
            resumption, previously trained on datamixes and durations cannot be
            changed. The duration of the current datamix must be greater than
            the saved timestamp. The dataset must be a StreamingDataset.
    """

    def __init__(
        self,
        train_config: dict[str, Any],
        schedule: list[dict[str, Any]],
        duration: Union[Time, str, int],
    ):
        # Ensure all duration units are in epochs or tokens and values are positive
        self._schedule = schedule
        if len(self._schedule) == 0:
            raise ValueError('The schedule must have at least one datamix.')
        first_datamix = {
            'duration': duration,
            'dataset': train_config['train_loader']['dataset'],
        }
        self._schedule.insert(0, first_datamix)
        for datamix in self._schedule:
            self._validate_datamix(datamix)

        self._schedule_index = 0
        self._train_loader_config: dict[str, Any] = train_config['train_loader']
        self._device_train_batch_size = train_config['device_train_batch_size']
        self._tokenizer = None

    def init(self, state: State, logger: Logger):
        del logger  # unused

        if not hasattr(state.model, 'tokenizer'):
            raise ValueError('state.model must have a tokenizer attribute.')
        self._tokenizer = state.model.tokenizer

    def before_load(self, state: State, logger: Logger):
        del logger  # unused

        # Ensure all duration units are the same as max_duration
        datamix_units = [datamix['duration'].unit for datamix in self._schedule]
        assert state.max_duration is not None, 'max_duration should have beeen set.'
        if any(state.max_duration.unit != unit for unit in datamix_units):
            raise ValueError((
                f'All durations in the schedule must have the same units as '
                f'the max_duration. Expected {state.max_duration.unit}, but '
                f'got {datamix_units}.'
            ))

        # Ensure schedule duration is equal to max_duration
        schedule_duration = Time(0, state.max_duration.unit)
        for datamix in self._schedule:
            assert isinstance(datamix['duration'], Time)
            schedule_duration += datamix['duration']
        if schedule_duration != state.max_duration:
            raise ValueError((
                'The sum of all durations in the schedule must be equal to the '
                'max_duration.'
            ))

        self._validate_dataloader(state.train_dataloader)

    def after_load(self, state: State, logger: Logger):
        del logger  # unused

        self._validate_dataloader(state.train_dataloader)

        # If checkpoint was saved before iteration was incremented, we need to increment it now
        duration = self._schedule[self._schedule_index]['duration']
        if ((
            duration.unit == TimeUnit.TOKEN and
            state.timestamp.token_in_iteration >= duration.value
        ) or (
            duration.unit == TimeUnit.EPOCH and
            state.timestamp.epoch_in_iteration >= duration.value
        )):
            log.warning((
                'The CurriculumLearning callback has detected that the '
                'previous run did not correctly increment the iteration.'
            ))
            self._schedule_index += 1
            state.timestamp = state.timestamp.to_next_iteration()

    def iteration_start(self, state: State, logger: Logger):
        # Swap the dataset if starting a new iteration that's not the original datamix
        if self._schedule_index > 0:
            # TODO: trainer._train_data_spec should be updated whenever the dataloader is updated
            # Dataloaders with the same prefix access the same shared memory
            # which is stale
            clean_stale_shared_memory()
            datamix = copy.deepcopy(self._schedule[self._schedule_index])
            train_loader_config = copy.deepcopy(self._train_loader_config)
            train_loader_config['dataset'].update(datamix['dataset'])
            data_spec = self._build_train_loader(
                train_loader_config=train_loader_config,
                logger=logger,
            )
            state.set_dataloader(
                dataloader=data_spec.dataloader,
                dataloader_label='train',
            )
            state.train_dataloader = state.dataloader
            self._validate_dataloader(state.train_dataloader)

        # Set the length of the new iteration
        state._iteration_length = self._schedule[self._schedule_index
                                                ]['duration']

    def iteration_end(self, state: State, logger: Logger):
        del state, logger  # unused

        self._schedule_index += 1

    def state_dict(self):
        return {
            'state':
                CurriculumLearningState(
                    schedule=self._schedule,
                    schedule_index=self._schedule_index,
                ),
        }

    def load_state_dict(self, state: dict[str, Any]):
        schedule = state['state'].schedule
        self._schedule_index = state['state'].schedule_index

        # Ensure that the schedule has not changed on previously trained datamixes
        for idx in range(self._schedule_index):
            if self._schedule[idx] != schedule[idx]:
                raise ValueError((
                    f'Previous datamixes must stay the same across ',
                    f'resumptions. Expected {schedule[idx]} but got ',
                    f'{self._schedule[idx]}',
                ))

        # Ensure that the datamix has not changed on the current datamix
        current_dataset = self._schedule[self._schedule_index]['dataset']
        saved_dataset = schedule[self._schedule_index]['dataset']
        if current_dataset != saved_dataset:
            raise ValueError((
                f'The current datamix must stay the same across resumptions. ',
                f'Expected {saved_dataset} but got {current_dataset}',
            ))

        # Ensure that the current datamix duration is in the correct units
        duration = self._schedule[self._schedule_index]['duration']
        if duration.unit != TimeUnit.TOKEN and duration.unit != TimeUnit.EPOCH:
            raise ValueError((
                f'Duration must be in terms of tokens or epochs, but got ',
                f'{duration.unit}.',
            ))

    def _build_train_loader(
        self,
        train_loader_config: dict[str, Any],
        logger: Logger,
    ) -> DataSpec:
        del logger  # unused

        from llmfoundry.data.dataloader import build_dataloader

        # Copied from scripts/train/train.py
        log.info(
            f'Building train loader in CurriculumLearning callback for dataset {self._schedule_index}',
        )
        assert self._tokenizer is not None
        try:
            return build_dataloader(
                train_loader_config,
                self._tokenizer,
                self._device_train_batch_size,
            )
        except BaseContextualError as e:
            e.location = TrainDataLoaderLocation
            raise e

    def _validate_dataloader(self, train_loader: Any):
        # Check if we are using a DataLoader and StreamingDataset
        if not isinstance(train_loader, DataLoader):
            raise ValueError(
                f'CurriculumLearning callback can only be used with a train ',
                f'dataloader of type DataLoader, but got {type(train_loader)}.',
            )
        dataset = train_loader.dataset
        if not isinstance(dataset, StreamingDataset):
            raise ValueError(
                f'CurriculumLearning callback only supports StreamingDataset ',
                f'because it requires loading and saving dataset state. ',
                f'Instead, got a dataset of type {type(dataset)}',
            )

    def _validate_datamix(self, datamix: dict[str, Any]):
        if 'duration' not in datamix:
            raise ValueError('Each datamix must have a duration.')
        datamix['duration'] = ensure_time(
            datamix['duration'],
            TimeUnit.EPOCH,
        )
        if datamix['duration'].value <= 0:
            raise ValueError('The duration must be positive.')
        if (
            datamix['duration'].unit != TimeUnit.EPOCH and
            datamix['duration'].unit != TimeUnit.TOKEN
        ):
            raise ValueError(
                'Schedules can only be defined in terms of epochs or tokens.',
            )

        if 'dataset' not in datamix:
            raise ValueError('Each datamix must have a dataset.')



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from composer.algorithms import (
    Alibi,
    GatedLinearUnits,
    GradientClipping,
    LowPrecisionLayerNorm,
)

from llmfoundry.registry import algorithms

algorithms.register('gradient_clipping', func=GradientClipping)
algorithms.register('alibi', func=Alibi)
algorithms.register('gated_linear_units', func=GatedLinearUnits)
algorithms.register('low_precision_layernorm', func=LowPrecisionLayerNorm)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Implements a Hugging Causal LM wrapped inside a :class:`.ComposerModel`."""

import logging
from typing import (
    Any,
    Optional,
    Union,
)

from transformers import (
    AutoModelForCausalLM,
    PreTrainedModel,
    PreTrainedTokenizerBase,
)
from transformers.models.auto.auto_factory import _BaseAutoModelClass

from llmfoundry.metrics import (
    DEFAULT_CAUSAL_LM_EVAL_METRICS,
    DEFAULT_CAUSAL_LM_TRAIN_METRICS,
)
from llmfoundry.models.hf.hf_base import BaseHuggingFaceModel

__all__ = ['ComposerHFCausalLM']

log = logging.getLogger(__name__)


class ComposerHFCausalLM(BaseHuggingFaceModel):
    """Configures a :class:`.HuggingFaceModel` around a Causal LM.

    Args:
        pretrained_model_name_or_path (str): The name of or local path to
            the HF Causal LM (e.g., `gpt2` to instantiate a GPT2LMHeadModel).
        config_overrides (dict, optional): An optional dictionary of keyword
            arguments that override the default configuration associated with
            cfg.pretrained_model_name_or_path.
        pretrained (bool): Whether to instantiate the model with pre-trained
            weights coming from cfg.pretrained_model_name_or_path. If ``True``,
            cfg.config_overrides must be compatible with the pre-trained weights.
        init_device ('cpu' | 'meta'): Which device, 'cpu' or 'meta', to
            initialize the model on. Currently, `meta` is only supported when
            cfg.pretrained is ``False``. Default: ``'cpu'``.
        peft_config (dict, optional): An optional dictionary of keyword arguments to be
            passed to the PeftConfig constructor. If provided, the model will be wrapped in a PeftModel.
        trust_remote_code (bool, optional): Whether to trust remote code when loading from Hugging Face
            Hub. Default: ``True``.
        use_auth_token (bool, optional): Whether to use the Hugging Face authentication token when
            loading from Hugging Face Hub. Default: ``False``.
        use_train_metrics (bool, optional): Whether to use training metrics. Default: ``True``.
        load_in_8bit (bool, optional): Whether to load the model in 8-bit mode. Default: ``False``.
        init_device (str, optional): Which device to initialize the model on. Default: ``'cpu'``.
        use_flash_attention_2 (bool, optional): Whether to use flash-attention 2. Default: ``False``.
        tokenizer (PreTrainedTokenizer): The tokenizer that the model will use.
    """

    model_cls: Union[_BaseAutoModelClass,
                     PreTrainedModel] = AutoModelForCausalLM
    default_train_metrics: tuple = tuple(DEFAULT_CAUSAL_LM_TRAIN_METRICS)
    default_eval_metrics: tuple = tuple(DEFAULT_CAUSAL_LM_EVAL_METRICS)

    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        pretrained_model_name_or_path: str,
        pretrained: bool = True,
        pretrained_lora_id_or_path: Optional[str] = None,
        trust_remote_code: bool = True,
        use_auth_token: bool = False,
        use_flash_attention_2: bool = False,
        load_in_8bit: bool = False,
        init_device: str = 'cpu',
        config_overrides: Optional[dict[str, Any]] = None,
        peft_config: Optional[dict[str, Any]] = None,
        use_train_metrics: bool = True,
        allow_embedding_resizing: bool = False,
        additional_train_metrics: Optional[list] = None,
        additional_eval_metrics: Optional[list] = None,
        should_save_peft_only: bool = True,
    ):
        super().__init__(
            pretrained_model_name_or_path,
            tokenizer=tokenizer,
            pretrained=pretrained,
            pretrained_lora_id_or_path=pretrained_lora_id_or_path,
            trust_remote_code=trust_remote_code,
            use_auth_token=use_auth_token,
            use_flash_attention_2=use_flash_attention_2,
            load_in_8bit=load_in_8bit,
            init_device=init_device,
            config_overrides=config_overrides,
            shift_labels=True,
            peft_config=peft_config,
            allow_embedding_resizing=allow_embedding_resizing,
            use_train_metrics=use_train_metrics,
            additional_train_metrics=additional_train_metrics,
            additional_eval_metrics=additional_eval_metrics,
            should_save_peft_only=should_save_peft_only,
        )



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.interfaces.callback_with_config import CallbackWithConfig

__all__ = [
    'CallbackWithConfig',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""A HuggingFace-style model configuration."""

import copy
import warnings
from typing import Any, Optional, Union

from transformers import PretrainedConfig

from llmfoundry.layers_registry import ffns_with_megablocks
from llmfoundry.models.layers.attention import (
    check_alibi_support,
    is_flash_v2_installed,
)
from llmfoundry.models.utils.config_defaults import (
    attn_config_defaults,
    fc_type_defaults,
    ffn_config_defaults,
    init_config_defaults,
)
from llmfoundry.utils.warnings import ExperimentalWarning


class MPTConfig(PretrainedConfig):
    model_type = 'mpt'

    def __init__(
        self,
        d_model: int = 2048,
        n_heads: int = 16,
        n_layers: int = 24,
        expansion_ratio: Union[int, float] = 4,
        max_seq_len: int = 2048,
        vocab_size: int = 50368,
        resid_pdrop: float = 0.0,
        emb_pdrop: float = 0.0,
        learned_pos_emb: bool = True,
        attn_config: Optional[dict] = None,
        ffn_config: Optional[dict] = None,
        init_device: str = 'cpu',
        logit_scale: Optional[Union[float, str]] = None,
        no_bias: bool = False,
        embedding_fraction: float = 1.0,
        norm_type: str = 'low_precision_layernorm',
        norm_eps: float = 1e-05,
        use_cache: bool = False,
        init_config: Optional[dict] = None,
        fc_type: Union[str, dict] = 'torch',
        tie_word_embeddings: bool = True,
        use_pad_tok_in_ffn: bool = True,
        block_overrides: Optional[dict[str, Any]] = None,
        final_logit_softcapping: Optional[float] = None,
        **kwargs: Any,
    ):
        """The MPT configuration class.

        Args:
            d_model (int): The size of the embedding dimension of the model.
            n_heads (int): The number of attention heads.
            n_layers (int): The number of layers in the model.
            expansion_ratio (Union[int, float]): The ratio of the up/down scale in the ffn.
            max_seq_len (int): The maximum sequence length of the model.
            vocab_size (int): The size of the vocabulary.
            resid_pdrop (float): The dropout probability applied to the attention output before combining with residual.
            emb_pdrop (float): The dropout probability for the embedding layer.
            learned_pos_emb (bool): Whether to use learned positional embeddings
            attn_config (Dict): A dictionary used to configure the model's attention module:
                attn_type (str): type of attention to use. Options: multihead_attention, multiquery_attention, grouped_query_attention
                attn_pdrop (float): The dropout probability for the attention layers.
                attn_impl (str): The attention implementation to use. One of 'torch' or 'flash'.
                qk_ln (bool): Whether to apply layer normalization to the queries and keys in the attention layer.
                qk_gn (bool): Whether to apply group normalization to the queries and keys in the attention layer.
                fused_qkv (bool): Whether to fuse the Wq, Wk, and Wv weight matrices in the attention layer. If True, the weights are fused into a single
                    Wqkv matrix, which can be faster for matmuls. If False, the weights are kept separate. Defaults to True.
                clip_qkv (Optional[float]): If not None, clip the queries, keys, and values in the attention layer to
                    this value.
                softmax_scale (Optional[float]): If not None, scale the softmax in the attention layer by this value. If None,
                    use the default scale of ``1/sqrt(d_keys)``.
                attn_uses_sequence_id (Optional[bool]): Whether to restrict attention to tokens that have the same sequence_id.
                    When the model is in `train` mode, this requires passing an extra `sequence_id` argument which indicates
                    which sub-sequence each token belongs to.
                    Defaults to ``False`` meaning any provided `sequence_id` will be ignored.
                sliding_window_size (int): Window size for sliding window local attention. Defaults to -1, which means no sliding window. Query at position i will only attend to keys between [i + seqlen_k - seqlen_q - window_size, i + seqlen_k - seqlen_q + window_size] inclusive. Only works for flash attention v2.3.0 or higher.
                alibi (bool): Whether to use the alibi bias instead of position embeddings.
                alibi_bias_max (int): The maximum value of the alibi bias.
                rope (bool): Whether to use rotary positional embeddings.
                rope_theta (int): The base frequency for rope.
                rope_impl (str): The implementation of rope to use. One of 'hf' (to use the implementation from https://github.com/huggingface/transformers/blob/main/src/transformers/models/llama/modeling_llama.py) or 'dail' (to use the implementation from https://github.com/Dao-AILab/flash-attention/blob/main/flash_attn/layers/rotary.py).
                rope_dail_config (Dict): The configuration for the dail implementation of rope.
                    type (str): The type of rotary position embedding to use. Options: 'original' (for https://arxiv.org/pdf/2104.09864.pdf), 'xpos' (for https://arxiv.org/pdf/2212.10554.pdf).
                    pos_idx_in_fp32 (bool): If True, the position indices [0, ..., seqlen - 1] are in fp32, otherwise they might be in lower precision. A consequence could be, for example, that bf16 rounds position 1995 to 2000, which leads to them having the same positional embedding.
                    xpos_scale_base (float): The scale base for XPos (if using XPos).
                rope_hf_config (Dict): A dictionary used to configure rope's scaling behavior (when scaling beyond the training length).
                    type (str): Can be one of 'no_scaling', 'linear', or 'dynamic'. 'no_scaling' uses the default implementation for rotary embeddings, 'linear' uses linear scaling as proposed by the Reddit user /u/kaiokendev, and 'dynamic' uses Dynamic NTK scaling as proposed by the Reddit users /u/bloc97 and /u/emozilla.
                    factor (float): Scaling factor to use if using 'linear' or 'dynamic' as rope_scaling.type.
                kv_n_heads (Optional[int]): For grouped_query_attention only, allow user to specify number of kv heads.
                kv_dim (Optional[int]): For cross-attention only, allow user to specify different input dimensions for kv projections.
            ffn_config (Dict): A dictionary used to configure the model's ffn module:
                ffn_type (str): type of ffn to use. Options: mptmlp, mptglu, te_ln_mlp
            init_device (str): The device to use for parameter initialization.
            logit_scale (Optional[Union[float, str]]): If not None, scale the logits by this value.
            no_bias (bool): Whether to use bias in all layers.
            embedding_fraction (float): The fraction to scale the gradients of the embedding layer by.
            norm_type (str): choose type of norm to use
            norm_eps (float): epsilon value for norm layer
            use_cache (bool): Whether or not the model should return the last key/values attentions
            init_config (Dict): A dictionary used to configure the model initialization:
                init_config.name: The parameter initialization scheme to use. Options: 'default_', 'baseline_',
                    'kaiming_uniform_', 'kaiming_normal_', 'neox_init_', 'small_init_', 'xavier_uniform_', or
                    'xavier_normal_'. These mimic the parameter initialization methods in PyTorch.
                init_div_is_residual (Union[int, float, str, bool]): Value to divide initial weights by if ``module._is_residual`` is True.
                emb_init_std (Optional[float]): The standard deviation of the normal distribution used to initialize the embedding layer.
                emb_init_uniform_lim (Optional[Union[Tuple[float, float], float]]): The lower and upper limits of the uniform distribution
                    used to initialize the embedding layer. Mutually exclusive with ``emb_init_std``.
                init_std (float): The standard deviation of the normal distribution used to initialize the model,
                    if using the baseline_ parameter initialization scheme.
                init_gain (float): The gain to use for parameter initialization with kaiming or xavier initialization schemes.
                fan_mode (str): The fan mode to use for parameter initialization with kaiming initialization schemes.
                init_nonlinearity (str): The nonlinearity to use for parameter initialization with kaiming initialization schemes.
                ---
                See llmfoundry.models.utils.param_init_fns.py for info on other param init config options
            fc_type (str | Dict): Choose fc layer implementation. Options: torch and te. te layers support fp8 when using H100 GPUs. Can
                also be a dictionary that specifies the fc layer name and any kwargs for the fc layer.
            tie_word_embeddings (bool): Whether to tie the input embedding and output layers.
            use_pad_tok_in_ffn (bool): Whether to forward the pad token in the feedforward networks.
            block_overrides: This allows for overriding default block configs for certain layers. This must contain `overrides` and `order`. `order` is a nested list which describes the order of the layers. For each kind of layer, specify the `overrides` in the overrides config (default refers to a layer that does not apply any overrides).
                To specify this model (https://research.character.ai/optimizing-inference/) , the following config will be needed:
                    block_overrides:
                        order:
                        - name: default
                        - repeat: 2
                          order:
                          - name: sliding_window_layer
                          - name: sliding_window_layer_reuse
                          - name: sliding_window_layer
                          - repeat: 2
                            name: sliding_window_layer_reuse
                          - name: reuse_kv_layer
                        overrides:
                            sliding_window_layer:
                                attn_config:
                                    sliding_window_size: 1024
                            sliding_window_layer_reuse:
                                attn_config:
                                    sliding_window_size: 1024
                                    reuse_kv_layer_idx: -1 # Relative index of the layer whose kv cache to reuse
                            reuse_kv_layer:
                                attn_config:
                                    reuse_kv_layer_idx: -6 # Relative index of the layer whose kv cache to reuse
            final_logit_softcapping (float | None): Softcapping threshold for final logit. Set to None to disable (default value None). Please see https://arxiv.org/pdf/2403.08295 for more details.
            kwargs (Any): Other relevant keyword arguments.
        """
        self.d_model = d_model
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.expansion_ratio = expansion_ratio
        if max_seq_len != int(max_seq_len):
            raise ValueError('max_seq_len must be an integer')
        self.max_seq_len = int(max_seq_len)
        self.vocab_size = vocab_size
        self.resid_pdrop = resid_pdrop
        self.emb_pdrop = emb_pdrop
        self.learned_pos_emb = learned_pos_emb
        self.attn_config = attn_config if attn_config is not None else copy.deepcopy(
            attn_config_defaults,
        )
        self.ffn_config = ffn_config if ffn_config is not None else copy.deepcopy(
            ffn_config_defaults,
        )
        self.init_device = init_device
        self.logit_scale = logit_scale
        self.no_bias = no_bias
        self.embedding_fraction = embedding_fraction
        self.norm_type = norm_type
        self.norm_eps = norm_eps
        self.use_cache = use_cache
        self.init_config = init_config if init_config is not None else copy.deepcopy(
            init_config_defaults,
        )

        if block_overrides is not None:
            self._validate_block_overrides(block_overrides)
        self.block_overrides = block_overrides
        self.final_logit_softcapping = final_logit_softcapping

        if isinstance(fc_type, str):
            fc_type = {'name': fc_type}
        self.fc_type = fc_type

        self.use_pad_tok_in_ffn = use_pad_tok_in_ffn

        if 'name' in kwargs:
            del kwargs['name']
        if 'loss_fn' in kwargs:
            del kwargs['loss_fn']
        if self.attn_config.get('alibi',
                                False) or self.attn_config.get('rope', False):
            self.learned_pos_emb = False
            warnings.warn(
                f'alibi or rope is turned on, setting `learned_pos_emb` to `False.`',
            )
        # tie_word_embeddings is set in Huggingface's PretrainedConfig __init__
        super().__init__(
            tie_word_embeddings=tie_word_embeddings,
            **kwargs,
        )

        self._validate_config()

    def _validate_block_overrides(self, block_overrides: dict[str, Any]):
        warnings.warn(ExperimentalWarning('block_overrides'))
        if 'order' not in block_overrides:
            raise ValueError('`order` should be defined in block_overrides',)
        if 'overrides' not in block_overrides:
            raise ValueError(
                '`overrides` should be defined in block_overrides',
            )
        if 'default' in block_overrides['overrides'].keys():
            raise ValueError('block overrides cannot be named "default".',)

    def _set_config_defaults(
        self,
        config: dict[str, Any],
        config_defaults: dict[str, Any],
    ) -> dict[str, Any]:
        # set config defaults
        for k, v in config_defaults.items():
            if k not in config:
                config[k] = v
            elif isinstance(v, dict):
                # recursively set default values for any sub-dicts
                config[k] = self._set_config_defaults(
                    config[k] if (config[k] is not None) else {},
                    v,
                )
        return config

    def validate_attention_config(self) -> None:
        if 'seq_parallel_world_size' in self.attn_config and self.attn_config[
            'seq_parallel_world_size'] is None:
            del self.attn_config['seq_parallel_world_size']
        if self.attn_config.get('seq_parallel_world_size', 1) > 1:
            raise NotImplementedError('Sequence Parallelism is not supported.')

    def _validate_config(self) -> None:
        # set config defaults
        self.attn_config = self._set_config_defaults(
            self.attn_config,
            attn_config_defaults,
        )
        self.ffn_config = self._set_config_defaults(
            self.ffn_config,
            ffn_config_defaults,
        )
        self.init_config = self._set_config_defaults(
            self.init_config,
            init_config_defaults,
        )
        self.fc_type = self._set_config_defaults(
            self.fc_type,
            fc_type_defaults,
        )

        if self.d_model % self.n_heads != 0:
            raise ValueError('d_model must be divisible by n_heads')
        if any(
            prob < 0 or prob > 1 for prob in
            [self.attn_config['attn_pdrop'], self.resid_pdrop, self.emb_pdrop]
        ):
            raise ValueError(
                "self.attn_config['attn_pdrop'], resid_pdrop, emb_pdrop are probabilities and must be between 0 and 1",
            )
        if self.attn_config['attn_impl'] not in ['torch', 'flash']:
            raise ValueError(
                f"Unknown attn_impl={self.attn_config['attn_impl']}",
            )
        if self.attn_config['alibi'] and not check_alibi_support(
            self.attn_config['attn_impl'],
        ):
            raise NotImplementedError(
                'alibi only implemented with torch and flash (v2.4.2 or higher) attention.',
            )
        if self.attn_config['attn_uses_sequence_id'] and not (
            self.attn_config['attn_impl'] == 'torch' or (
                self.attn_config['attn_impl'] == 'flash' and
                is_flash_v2_installed(v2_version='v2.1.2')
            )
        ):
            raise NotImplementedError(
                'attn_uses_sequence_id only implemented with torch and flash (v2.1.2 or higher) attention.',
            )
        if self.attn_config['rope'] and (
            self.attn_config['rope_impl'] not in ['dail', 'hf']
        ):
            raise ValueError(
                'If rope is being used then rope_impl should be either "dail", or "hf".',
            )
        if self.attn_config['rope'] and (
            self.attn_config['rope_impl'] == 'hf'
        ) and self.attn_config['rope_hf_config']['type'] not in [
            'no_scaling',
            'linear',
            'dynamic',
            'llama3',
        ]:
            raise ValueError(
                'If using hf implementation of rope, the type should be one of "no_scaling", "linear" or "dynamic".',
            )
        if self.attn_config['rope'] and (
            self.attn_config['rope_impl'] == 'dail'
        ):
            if self.attn_config['rope_dail_config']['type'] not in [
                'original',
                'xpos',
            ]:
                raise ValueError(
                    'If using the dail implementation of rope, the type should be one of "original" or "xpos".',
                )
            if not is_flash_v2_installed(v2_version='2.0.1'):
                raise ImportError(
                    'If using the dail implementation of rope, the flash_attn library v2.0.1 or higher must be installed. Please check the instructions at https://github.com/mosaicml/llm-foundry/blob/main/TUTORIAL.md#what-kinds-of-positional-embeddings-does-llm-foundry-support',
                )
        if self.attn_config['sliding_window_size'] != -1 and self.attn_config[
            'attn_impl'
        ] == 'flash' and not is_flash_v2_installed(v2_version='v2.3.0',):
            raise NotImplementedError(
                'sliding window attention only implemented for torch attention and flash attention (v2.3.0 or higher).',
            )
        if self.attn_config['attn_logit_softcapping'] is not None:
            if self.attn_config['attn_logit_softcapping'] <= 0:
                raise ValueError(
                    'Attention attn_logit_softcapping should be positive.',
                )
            if self.attn_config[
                'attn_impl'
            ] == 'flash' and not is_flash_v2_installed(v2_version='v2.6.2',):
                raise NotImplementedError(
                    'Attention attn_logit_softcapping is only implemented with torch attention or flash attention v2.6.2 (or higher).',
                )
        if self.attn_config['kv_dim'] is not None and self.attn_config[
            'fused_qkv']:
            raise ValueError(
                'fused_qkv should be False when "kv_dim" is specified.',
            )
        if self.embedding_fraction > 1 or self.embedding_fraction <= 0:
            raise ValueError(
                'model.embedding_fraction must be between 0 (exclusive) and 1 (inclusive)!',
            )
        if isinstance(
            self.logit_scale,
            str,
        ) and self.logit_scale != 'inv_sqrt_d_model':
            raise ValueError(
                f"{self.logit_scale=} is not recognized as an option; use numeric value or 'inv_sqrt_d_model'.",
            )
        if self.init_config.get('name', None) is None:
            raise ValueError(f"{self.init_config=} 'name' needs to be set.")
        if not (
            self.learned_pos_emb or self.attn_config['alibi'] or
            self.attn_config['rope']
        ):
            warnings.warn(
                f'Positional information not being provided to the model using either learned_pos_emb or alibi or rope.',
            )
        if self.fc_type['name'] == 'te' or self.ffn_config['ffn_type'
                                                          ] == 'te_ln_mlp':
            try:
                import transformer_engine.pytorch as te
                del te  # unused
            except:
                raise ImportError(
                    'TransformerEngine import failed. `fc_type: te` requires TransformerEngine be installed, ',
                    'e.g. pip install transformer-engine[pytorch]',
                )

        self.ffn_config['fc_type'] = self.fc_type
        if self.ffn_config['ffn_type'] == 'mptgeglu':
            raise ValueError(
                'API CHANGE: `ffn_type=="mptgeglu"` changed to `ffn_type=="mptglu"`. '
                +
                'See [#829](https://github.com/mosaicml/llm-foundry/pull/829) for details.',
            )
        elif self.ffn_config['ffn_type'] in ffns_with_megablocks:
            self.ffn_config['return_bias'] = False
        elif self.ffn_config['ffn_type'] == 'te_ln_mlp':
            self.ffn_config['bias'] = not self.no_bias
            if 'ffn_act_fn' in self.ffn_config.keys():
                raise ValueError(
                    f'Transformer Engine block does not support custom activation functions.',
                )
        if not self.use_pad_tok_in_ffn:
            try:
                from flash_attn.bert_padding import unpad_input, pad_input  # type: ignore # yapf: disable # isort: skip
            except:
                raise ImportError(
                    'In order to set `use_pad_tok_in_ffn=False`, please install flash-attn==1.0.9 or flash-attn==2.3.6',
                )

        self.validate_attention_config()

    @property
    def allowed_block_overrides(self):
        return {
            'attn_config': {
                'sliding_window_size': None,
                'reuse_kv_layer_idx': None,
            },
        }



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Enable curriculum learning by resuming with a different dataset.

This callback is currently experimental. The API may change without warning in
the future.
"""

import logging
from dataclasses import dataclass

from composer.core import State
from composer.loggers import Logger
from streaming import StreamingDataset
from torch.utils.data import DataLoader

from llmfoundry.interfaces import CallbackWithConfig
from llmfoundry.utils.warnings import experimental_class

log = logging.getLogger(__name__)

__all__ = ['DatasetSwap']


@dataclass
class DatasetSwapStateDict:
    dataset_index: int
    all_dataset_configs: list


@experimental_class('DatasetSwap callback')
class DatasetSwap(CallbackWithConfig):
    """Starts an epoch with a different dataset when resuming from a checkpoint.

    Args:
        train_config (Dict): The configuration of the dataset currently
            being used. Note that this is the full train config and must
            contain the 'train_loader' key.
        dataset_index (int): The index of the dataset currently being used.
    """

    def __init__(self, train_config: dict, dataset_index: int):
        self.dataset_index = dataset_index
        self.saved_dataset_index = 0
        self.all_dataset_configs = []
        self.current_dataset_state = {}
        # The current dataset config is resolved and passed in train.py
        self.current_dataset_config = train_config['train_loader']

    def before_load(self, state: State, logger: Logger):
        del logger

        # Save the current dataset state so we can restore it correctly
        # if we are resuming with a new dataset.
        train_loader = state.train_dataloader
        # Check if we are using a DataLoader and StreamingDataset
        if not isinstance(train_loader, DataLoader):
            raise ValueError(
                f'CurriculumLearning callback can only be used with a train ',
                f'dataloader of type DataLoader, but got {type(train_loader)}.',
            )
        dataset = train_loader.dataset
        if not isinstance(dataset, StreamingDataset):
            raise ValueError(
                f'CurriculumLearning callback only supports StreamingDataset ',
                f'because it requires loading and saving dataset state. ',
                f'Instead, got a dataset of type {type(dataset)}',
            )
        assert isinstance(dataset, StreamingDataset)
        # Save the current dataset state so we can restore it if needed.
        self.current_dataset_state = dataset.state_dict(  # type: ignore
            num_samples=0, from_beginning=False)

    def after_load(self, state: State, logger: Logger):
        del logger

        # As saved_dataset_index is loaded from state_dict, this only runs when
        # a user explicitly increments the dataset_index and not on any other
        # resumption, including autoresume.
        train_loader = state._train_dataloader
        assert isinstance(
            train_loader,
            DataLoader,
        ), 'CurriculumLearning callback requires a DataLoader.'
        dataset = train_loader.dataset
        assert isinstance(
            dataset,
            StreamingDataset,
        ), 'CurriculumLearning callback requires a StreamingDataset.'
        if self.saved_dataset_index < self.dataset_index:
            # Ignore the dataset state that was read in from the checkpoint, and
            # replace with the new dataset state. This preserves resumption info.
            if self.current_dataset_state['epoch'] < 0:
                # Make sure the epoch in the loaded state dict is not negative.
                # Since `__iter__` has not yet been called on the dataset, the
                # epoch index in the dataset will still be -1. We need to ensure
                # that we set the epoch correctly to 0 in this case.
                self.current_dataset_state['epoch'] = 0
            dataset.load_state_dict(  # type: ignore
                self.current_dataset_state)
            # Start a new epoch since we are using a new dataset.
            # This will also reset the sample_in_epoch written to checkpoint,
            # making sure that subsequent resumptions proceed correctly.
            state.timestamp = state.timestamp.to_next_epoch()
            # Append the new dataset config to the list of all dataset configs.
            self.all_dataset_configs.append(self.current_dataset_config)
        elif self.dataset_index == 0 and len(self.all_dataset_configs) == 0:
            # Make sure to track our current dataset config if we are just starting training.
            self.all_dataset_configs.append(self.current_dataset_config)

    def state_dict(self):
        return {
            'callback_state':
                DatasetSwapStateDict(
                    dataset_index=self.dataset_index,
                    all_dataset_configs=self.all_dataset_configs,
                ),
        }

    def load_state_dict(self, state: dict[str, DatasetSwapStateDict]):
        _dummy_obj = DatasetSwapStateDict(
            dataset_index=0,
            all_dataset_configs=[],
        )
        _state_obj = state.get('callback_state', _dummy_obj)
        self.saved_dataset_index = getattr(_state_obj, 'dataset_index')
        self.all_dataset_configs = getattr(_state_obj, 'all_dataset_configs')
